#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Virginia Gradebook views.

$Id:$
"""
import datetime
import csv
import tempfile

from zope.component import queryUtility
from zope.traversing.api import getParents
from zope.security.proxy import removeSecurityProxy

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.course.interfaces import ISection, ISectionContainer
from schooltool.term.interfaces import IDateManager
from schooltool.basicperson.interfaces import IDemographics

from cando.gradebook.gradebook import _getEvaluation
from cando.virginia.competency import getCoreCompetencies
from cando.virginia.competency import buildCompetencyDict
from cando.virginia.competency import getFlatCompetencies
from cando.iep.interfaces import IStudentIEP
from cando.gradebook.interfaces import ICompetencyGradebook


def formatPercent(value):
    return '%2.0f%%' % (value + 0.0000000001)


class BaseReportView(object):

    def getStudentReportComps(self, student, section):
        return getFlatCompetencies(self.context.comprefDict)

    def __init__(self, context, request):
        self.context = removeSecurityProxy(context)
        self.request = request
        self.studentAggregate = {}
        self.averageNumberCompetencies = 0
        self.averagePercentGradedCompetent = 0.0
        self.averageGrade = 0.0
        self.averagePoints = 0.0
        self.num90Competent = self.num80Competent = self.num70Competent = \
            self.num60Competent = self.num100Grade = self.num90Grade = \
            self.num80Grade = self.num70Grade = self.num60Grade = \
            self.numLess60Grade = 0
        numGradedStudents = 0
        numStudents = 0
        section = self.section
        for student in self.getSectionStudents(section):
            numStudents += 1
            numCompetent = 0
            numGraded = 0
            numStudentComps = 0
            totalPoints = 0
            for comp in self.getStudentReportComps(student, section):
                numStudentComps += 1
                ev = _getEvaluation(student, comp)
                if ev is not None and \
                   self.scoresystem.getNumericalValue(ev.value) is not None:
                    numGraded += 1
                    totalPoints += ev.value.value
                    if ev.value.value >= self.scoresystem.minPassingScore.value:
                        numCompetent += 1
            if numGraded:
                numGradedStudents += 1
                aggregate = self.studentAggregate.setdefault(student, {})
                aggregate['numGraded'] = numGraded
                aggregate['totalPoints'] = totalPoints
                aggregate['numCompetent'] = numCompetent
                aggregate['percentGradedCompetent'] = float(100 * numCompetent) / float(numGraded)
                aggregate['numCompsTimes3'] = float(100 * totalPoints) / float(3 * numGraded)
                self.averageNumberCompetencies += aggregate['numCompetent']
                self.averagePercentGradedCompetent += aggregate['percentGradedCompetent']
                self.averageGrade += aggregate['numCompsTimes3']
                self.averagePoints += aggregate['totalPoints']
                if aggregate['percentGradedCompetent'] >= 90.0:
                    self.num90Competent += 1
                if aggregate['percentGradedCompetent'] >= 80.0:
                    self.num80Competent += 1
                if aggregate['percentGradedCompetent'] >= 70.0:
                    self.num70Competent += 1
                if aggregate['percentGradedCompetent'] >= 60.0:
                    self.num60Competent += 1
                if aggregate['numCompsTimes3'] >= 100.0:
                    self.num100Grade += 1
                if aggregate['numCompsTimes3'] >= 90.0:
                    self.num90Grade += 1
                if aggregate['numCompsTimes3'] >= 80.0:
                    self.num80Grade += 1
                if aggregate['numCompsTimes3'] >= 70.0:
                    self.num70Grade += 1
                if aggregate['numCompsTimes3'] >= 60.0:
                    self.num60Grade += 1
                if aggregate['numCompsTimes3'] < 60.0:
                    self.numLess60Grade += 1
        self.numGradedStudents = numGradedStudents
        self.percentTotalStudents = 0.0
        if numStudents:
            self.percentTotalStudents = float(100 * self.numGradedStudents) / float(numStudents)
        if self.numGradedStudents:
            self.averageNumberCompetencies = self.averageNumberCompetencies / \
                self.numGradedStudents
            self.averagePercentGradedCompetent = self.averagePercentGradedCompetent / \
                float(self.numGradedStudents)
            self.averageGrade = self.averageGrade / \
                float(self.numGradedStudents)
            self.averagePoints = self.averagePoints / \
                float(self.numGradedStudents)
        self.percent90Competent = self.calcPercent(self.num90Competent)
        self.percent80Competent = self.calcPercent(self.num80Competent)
        self.percent70Competent = self.calcPercent(self.num70Competent)
        self.percent60Competent = self.calcPercent(self.num60Competent)
        self.percent100Grade = self.calcPercent(self.num100Grade)
        self.percent90Grade = self.calcPercent(self.num90Grade)
        self.percent80Grade = self.calcPercent(self.num80Grade)
        self.percent70Grade = self.calcPercent(self.num70Grade)
        self.percent60Grade = self.calcPercent(self.num60Grade)
        self.percentLess60Grade = self.calcPercent(self.numLess60Grade)

    @property
    def section(self):
        for obj in reversed(getParents(self.context)):
            if ISection.providedBy(obj):
                return obj

    @property
    def scoresystem(self):
        return ISchoolToolApplication(None)['scoresystem']

    def calcPercent(self, value):
        if self.numGradedStudents:
            return float(100 * value) / float(self.numGradedStudents)
        else:
            return 0.0

    def getDivision(self, course):
        comps = getCoreCompetencies(course)
        try:
            return comps[0].__parent__.__parent__.__parent__.title
        except (IndexError, AttributeError,):
            return ''

    def getDivisionsTitles(self, section):
        return ", ".join([self.getDivision(course)
                          for course in section.courses
                          if self.getDivision(course)])

    def getCoursesTitles(self, section):
        return ", ".join([course.title for course in section.courses])

    def getInstructorsNames(self, section):
        return ", ".join(["%s %s" % (instructor.first_name,
                                     instructor.last_name,)
                          for instructor in section.instructors])

    def getFirstInstructorLastName(self, section):
        try:
            first_instructor = list(section.instructors)[0]
            return first_instructor.last_name
        except (IndexError, AttributeError,):
            return ""

    def getCoreCompetencies(self, section):
        result = {}
        for comp in getCoreCompetencies(section):
            result[comp] = None
        for course in list(section.courses):
            for comp in getCoreCompetencies(course):
                result[comp] = None
        return result.keys()

    def sectionCoreRequiredCompetencies(self, section):
        return [comp for comp in self.getCoreCompetencies(section)
                if comp.optionalFlag == "required"]

    def coreRequiredCompetenciesFor(self, student, section):
        student_iep = IStudentIEP(student)
        section_iep_comps = student_iep.getIEPCompetencies(section)
        if section_iep_comps is None:
            return self.sectionCoreRequiredCompetencies(section)
        return [comp for comp in self.sectionCoreRequiredCompetencies(section)
                if comp.globalId not in section_iep_comps]

    def getSections(self):
        term = queryUtility(IDateManager).current_term
        for section in ISectionContainer(term).values():
            yield section

    def getSectionStudents(self, section):
        for student in section.members:
            yield student

    def getCoursesStateCodes(self, section):
        return ", ".join([course.government_id for course in section.courses
                          if course.government_id is not None])

    def getCoursesLocalCodes(self, section):
        section_code = ", ".join([course.course_id
                                  for course in section.courses
                                  if course.course_id is not None])
        if section.description:
            if len(section.description.split(",")) == 2:
                section_code = section.description.split(",")[1].strip()
        return section_code


class VirginiaReportView(BaseReportView):
    
    def getStudents(self):
        results = []
        for student in sorted(self.studentAggregate, key=lambda x: x.title):
            aggregate = self.studentAggregate[student]
            dict = {'title': student.title,
                    'percentGradedCompetent':
                        formatPercent(aggregate['percentGradedCompetent']),
                    'numCompetent': aggregate['numCompetent'],
                    'numGraded': aggregate['numGraded']}
            results.append(dict)
        return results

    def getAggregate(self):
        return {'numStudents': self.numGradedStudents,
                'percentTotalStudents':
                    formatPercent(self.percentTotalStudents),
                'num90Competent': self.num90Competent,
                'percent90Competent': formatPercent(self.percent90Competent),
                'num80Competent': self.num80Competent,
                'percent80Competent': formatPercent(self.percent80Competent),
                'num70Competent': self.num70Competent,
                'percent70Competent': formatPercent(self.percent70Competent),
                'num60Competent': self.num60Competent,
                'percent60Competent': formatPercent(self.percent60Competent),
                'averagePercentGradedCompetent':
                    formatPercent(self.averagePercentGradedCompetent),
                'averageNumberCompetencies': self.averageNumberCompetencies}


class TeacherReportView(BaseReportView):

    def getStudents(self):
        results = []
        for student in sorted(self.studentAggregate, key=lambda x: x.title):
            aggregate = self.studentAggregate[student]
            dict = {'title': student.title,
                    'totalPoints': aggregate['totalPoints'],
                    'percentGradedCompetent':
                        formatPercent(aggregate['percentGradedCompetent']),
                    'ratedGreater3': aggregate['numCompetent'],
                    'evaluated': aggregate['numGraded'],
                    'evaluatedTimes3': 3 * aggregate['numGraded'],
                    'numCompsTimes3':
                        formatPercent(aggregate['numCompsTimes3'])}
            results.append(dict)
        return results

    def getAggregate(self):
        return {'averageGrade': formatPercent(self.averageGrade),
                'averagePoints': '%2.2f' % self.averagePoints,
                'totalStudents': self.numGradedStudents,
                'num100Grade': self.num100Grade,
                'percent100Grade': formatPercent(self.percent100Grade),
                'num90Grade': self.num90Grade,
                'percent90Grade': formatPercent(self.percent90Grade),
                'num80Grade': self.num80Grade,
                'percent80Grade': formatPercent(self.percent80Grade),
                'num70Grade': self.num70Grade,
                'percent70Grade': formatPercent(self.percent70Grade),
                'num60Grade': self.num60Grade,
                'percent60Grade': formatPercent(self.percent60Grade),
                'numLess60Grade': self.numLess60Grade,
                'percentLess60Grade': formatPercent(self.percentLess60Grade)}


class SectionReportView(BaseReportView):

    def __init__(self, context, request):
        self.context = removeSecurityProxy(context)
        self.request = request
        comps = getFlatCompetencies(self.context.comprefDict)
        self.all_total = len(comps)
        rows = []
        attainment_labels = {100: "90-100%",
                             90: "80-89%",
                             80: "70-79%",
                             70: "60-69%",
                             60: "<= 59%"}
        rows_attainment = {100: 0,
                           90: 0,
                           80: 0,
                           70: 0,
                           60: 0}
        rows_attainment80 = {u"80% or above": 0,
                             u"Below 80%": 0}
        self.student_total = len(self.section.members)
        section = self.section
        for student in self.getSectionStudents(section):
            demographics = IDemographics(student)
            student_core_required = self.coreRequiredCompetenciesFor(student, section)
            student_core_required_total = len(student_core_required)
            info = {"first_name": student.first_name,
                    "last_name": student.last_name,
                    "state_report_id": demographics.get("state_report_id", ""),
                    "required_evaluated": 0,
                    "required_competent": 0,
                    "all_evaluated": 0,
                    "all_competent": 0,
                    "student_core_required_total": student_core_required_total}
            for comp in comps:
                ev = _getEvaluation(student, comp)
                if ev is not None and \
                   self.scoresystem.getNumericalValue(ev.value) is not None:
                    info["all_evaluated"] += 1
                    if comp in student_core_required:
                        info["required_evaluated"] += 1
                    value = ev.value.value
                    # XXX: Hard coded 2 asked by Welsh
                    if value >= 2: 
                        info["all_competent"] += 1
                        if comp in student_core_required:
                            info["required_competent"] += 1
            try:
                required_rating = float(info["required_competent"] * 100) / \
                                  float(info["required_evaluated"])
            except (ZeroDivisionError,):
                required_rating = 0.0
            info["required_rating"] = formatPercent(required_rating)
            try:
                required_completed = float(info["required_competent"] * 100) / \
                                     float(student_core_required_total)
            except (ZeroDivisionError,):
                required_completed = 0.0
            info["required_completed"] = formatPercent(required_completed)
            try:
                all_rating = float(info["all_competent"] * 100) / \
                             float(info["all_evaluated"])
            except (ZeroDivisionError,):
                all_rating = 0.0
            if required_completed >= 90:
                rows_attainment[100] += 1
            elif required_completed >= 80:
                rows_attainment[90] += 1
            elif required_completed >= 70:
                rows_attainment[80] += 1
            elif required_completed >= 60:
                rows_attainment[70] += 1
            else:
                rows_attainment[60] += 1
            if required_completed >= 80:
                rows_attainment80[u"80% or above"] += 1
            else:
                rows_attainment80[u"Below 80%"] += 1                
            info["all_rating"] = formatPercent(all_rating)
            rows.append(info)
        self.rows = sorted(rows,
                           key=lambda x:(x["last_name"], x["first_name"]))
        self.rows_attainment = []
        self.rows_attainment80 = []
        for key in sorted(rows_attainment, reverse=True):
            info = {"scale": attainment_labels[key],
                    "student_count": rows_attainment[key]}
            try:
                student_percent = "%.2f" % (float(rows_attainment[key] * 100) / float(self.student_total),)
            except (ZeroDivisionError,):
                student_percent = "0.00"
            info["student_percent"] = student_percent
            self.rows_attainment.append(info)
        for key in sorted(rows_attainment80):
            info = {"scale": key,
                    "student_count": rows_attainment80[key]}
            try:
                student_percent = "%.2f" % (float(rows_attainment80[key] * 100) / float(self.student_total),)
            except (ZeroDivisionError,):
                student_percent = "0.00"
            info["student_percent"] = student_percent
            self.rows_attainment80.append(info)
        self.section_local_courses = []
        title = u"No section description"
        self.courses = list(self.section.courses)
        section_code = ", ".join([course.course_id
                                  for course in self.courses
                                  if course.course_id is not None])
        self.section_local_courses = self.courses
        if self.section.description:
            if len(self.section.description.split(",")) == 2:
                title = self.section.description.split(",")[0].strip()
                section_code = self.section.description.split(",")[1].strip()
                self.section_local_courses = []
                for course in self.courses:
                    if section_code in course.course_id:
                        self.section_local_courses.append(course)
            else:
                title = self.section.description
        self.section_title = title
        self.section_code = section_code
        self.instructors = self.getInstructorsNames(self.section)


class BaseCompletionReportView(BaseReportView):

    def __init__(self, context, request):
        self.context = removeSecurityProxy(context)
        self.request = request
        self.date = datetime.datetime.today().strftime("%b %d, %Y %I:%M%p")
        self.rows = []
        for section in self.getSections():
            averageNumberGraded = 0
            averagePercentGradedCompetent = 0.0
            averagePercentAllCompetent = 0.0
            average80Competent = 0.0
            numGradedStudents = 0
            numStudents = 0
            section_report_competencies =  self.getSectionReportComps(section)
            for student in self.getSectionStudents(section):
                numStudents += 1
                numCompetent = 0
                numGraded = 0
                numStudentComps = 0
                totalPoints = 0
                for comp in self.getStudentReportComps(student, section, section_report_competencies):
                    numStudentComps += 1
                    ev = _getEvaluation(student, comp)
                    if ev is not None and \
                       self.scoresystem.getNumericalValue(ev.value) is not None:
                        numGraded += 1
                        totalPoints += ev.value.value
                        if ev.value.value >= self.scoresystem.minPassingScore.value:
                            numCompetent +=1
                if numGraded:
                    numGradedStudents += 1
                    percentGradedCompetent = float(100 * numCompetent) / float(numGraded)

                    averageNumberGraded += numGraded
                    averagePercentGradedCompetent += percentGradedCompetent
                    averagePercentAllCompetent += float(100 * numCompetent) / float(numStudentComps)
                    if percentGradedCompetent >= 80.0:
                        average80Competent += 1
            if numGradedStudents:
                averageNumberGraded = averageNumberGraded / numGradedStudents
                averagePercentGradedCompetent = averagePercentGradedCompetent / float(numGradedStudents)
                averagePercentAllCompetent = averagePercentAllCompetent / float(numGradedStudents)
                average80Competent = float(100 * average80Competent) / float(numGradedStudents)
            row = {
                "division": self.getDivisionsTitles(section),
                "course": self.getCoursesTitles(section),
                "section": section.title,
                "instructor": self.getInstructorsNames(section),
                "last_name": self.getFirstInstructorLastName(section),
                "total": len(section_report_competencies),
                "evaluated": averageNumberGraded,
                "all_competent": formatPercent(averagePercentAllCompetent).strip(),
                "marked_competent": formatPercent(averagePercentGradedCompetent).strip(),
                "marked_competent80": formatPercent(average80Competent).strip(),
                }
            self.rows.append(row)
        sortingKey = lambda x:(x['division'],x['course'],x['last_name'])
        self.sortBy = u"Program Area"
        if self.request.get("sortBy") == 'Instructor':
            sortingKey = lambda x:(x['last_name'],x['division'],x['course'])
            self.sortBy = u"Instructor"
        self.rows = sorted(self.rows, key=sortingKey)


class CompletionReportView(BaseCompletionReportView):

    def getSectionReportComps(self, section):
        comprefDict = buildCompetencyDict(section)
        return getFlatCompetencies(comprefDict)

    def getStudentReportComps(self, student, section, section_report_competencies):
        return section_report_competencies


class RequiredCompletionReportView(BaseCompletionReportView):

    def getSectionReportComps(self, section):
        return self.sectionCoreRequiredCompetencies(section)

    def getStudentReportComps(self, student, section, section_report_competencies):
        return self.coreRequiredCompetenciesFor(student, section, section_report_competencies)
 
    def coreRequiredCompetenciesFor(self, student, section, section_report_competencies):
        student_iep = IStudentIEP(student)
        section_iep_comps = student_iep.getIEPCompetencies(section)
        if section_iep_comps is None:
            return section_report_competencies
        return [comp for comp in section_report_competencies
                if comp.globalId not in section_iep_comps]


class SchoolInfoSysExportView(BaseReportView):
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.rows = []
        self.date = datetime.datetime.today().strftime("%b %d, %Y %I:%M%p")
        for section in self.getSections():
            state_course_code = self.getCoursesStateCodes(section)
            local_course_code = self.getCoursesLocalCodes(section)            
            # loop though comps, students, aggregating values
            for student in self.getSectionStudents(section):
                demographics = IDemographics(student)
                testing_number = demographics.get('state_report_id', '')
                student_core_required = self.coreRequiredCompetenciesFor(student, section)
                num_core_required_comps = len(student_core_required)
                total_iep_comps = 0
                student_iep = IStudentIEP(student)
                section_iep_comps = student_iep.getIEPCompetencies(section)
                if section_iep_comps is not None:
                    total_iep_comps = len(section_iep_comps)
                section_fields = ['', '', '', '']
                for i, field in enumerate(section.title.split(',')):
                    section_fields.insert(i, field.strip())
                row = [testing_number, student.username,
                       state_course_code, local_course_code]
                row.extend(section_fields[:4])
                row.extend([num_core_required_comps, total_iep_comps])
                numCoreRequiredCompetent = 0
                required_evaluated = 0
                for comp in student_core_required:
                    ev = _getEvaluation(student, comp)
                    if ev is None or \
                       self.scoresystem.getNumericalValue(ev.value) is None:
                        continue
                    required_evaluated += 1
                    value = ev.value.value
                    if value >= self.scoresystem.minPassingScore.value:
                        numCoreRequiredCompetent += 1
                if not required_evaluated:
                    continue
                row.append(numCoreRequiredCompetent)
                self.rows.append(row)
        self.rows = sorted(self.rows)

    def __call__(self):
        result = tempfile.TemporaryFile()
        writer = csv.writer(result)
        writer.writerows(self.rows)
        self.request.response.setHeader("Content-Type", "text/csv")
        return result


class SchoolYearArchiveView(BaseReportView):

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.rows = []
        for section in self.getSections():
            section_report_competencies = self.sectionCoreRequiredCompetencies(section) 
            state_course_code = self.getCoursesStateCodes(section)
            local_course_code = self.getCoursesLocalCodes(section)
            gradebook = ICompetencyGradebook(section)
            comps = getFlatCompetencies(gradebook.comprefDict)
            all_total = len(comps)
            # loop though comps, students, aggregating values
            for student in self.getSectionStudents(section):
                demographics = IDemographics(student)
                testing_number = demographics.get('state_report_id', '')
                student_core_required = self.coreRequiredCompetenciesFor(student, section,
                                                                         section_report_competencies)
                num_core_required_comps = len(student_core_required)
                section_fields = ['', '', '', '']
                for i, field in enumerate(section.title.split(',')):
                    section_fields.insert(i, field.strip())
                row = [student.username, testing_number,
                       state_course_code, local_course_code]
                row.extend(section_fields[:4])
                row.append(num_core_required_comps)
                required_evaluated = 0
                required_competent = 0
                all_evaluated = 0
                all_competent = 0
                for comp in comps:
                    ev = _getEvaluation(student, comp)
                    if ev is not None and \
                       self.scoresystem.getNumericalValue(ev.value) is not None:
                        all_evaluated += 1
                        if comp in student_core_required:
                            required_evaluated += 1
                        value = ev.value.value
                        # XXX: Hard coded 2 asked by Welsh
                        if value >= 2: 
                            all_competent += 1
                            if comp in student_core_required:
                                required_competent += 1
                try:
                    required_rating = float(required_competent * 100) / \
                                      float(required_evaluated)
                except (ZeroDivisionError,):
                    required_rating = 0.0
                required_rating = formatPercent(required_rating)
                try:
                    required_completed = float(required_competent * 100) / \
                                         float(num_core_required_comps)
                except (ZeroDivisionError,):
                    required_completed = 0.0
                required_completed = formatPercent(required_completed)
                try:
                    all_rating = float(all_competent * 100) / \
                                 float(all_evaluated)
                except (ZeroDivisionError,):
                    all_rating = 0.0
                all_rating = formatPercent(all_rating)
                row.extend([required_evaluated, required_competent, required_rating,
                            required_completed, all_total, all_evaluated, all_competent,
                            all_rating])
                self.rows.append(row)

    def __call__(self):
        result = tempfile.TemporaryFile()
        writer = csv.writer(result)
        writer.writerows(self.rows)
        self.request.response.setHeader("Content-Type", "text/csv")
        return result

    def coreRequiredCompetenciesFor(self, student, section, section_report_competencies):
        student_iep = IStudentIEP(student)
        section_iep_comps = student_iep.getIEPCompetencies(section)
        if section_iep_comps is None:
            return section_report_competencies
        return [comp for comp in section_report_competencies
                if comp.globalId not in section_iep_comps]
