#
# CanDo - Web based competency tracking system.
# Copyright (C) 2007 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
import zope.schema
from zope.publisher.browser import BrowserView
from zope.component import getMultiAdapter
from zope.publisher.interfaces import IPublishTraverse
from zope.traversing.api import getName, getParent
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
from zope.traversing.browser import absoluteURL
from zope.viewlet.viewlet import JavaScriptViewlet
from zope.session.interfaces import ISession

from z3c.form import form, field
from z3c.form.interfaces import IWidgets

from schooltool.requirement.requirement import getRequirementKey

from cando.gradebook import interfaces, gradebook
from cando.virginia.interfaces import ICompetencies, ICompetencyReference

SESSION_KEY = 'cando.gradebook.browser.section'
sectionGradebookJavaScriptViewlet = JavaScriptViewlet('section-gradebook.js')

class SessionProperty(object):

    def __init__(self, name, default=None):
        self.name = name
        self.default = default

    def __get__(self, inst, klass):
        session = ISession(inst.request)[SESSION_KEY]
        return session.get(self.name, self.default)

    def __set__(self, inst, value):
        session = ISession(inst.request)[SESSION_KEY]
        session[self.name] = value


class BaseSectionGradebookView(BrowserView):
    """Base view that provides some nice features."""

    def __init__(self, context, request):
        super(BaseSectionGradebookView, self).__init__(context, request)

        # find a parent object that isn't a gradebook for reference
        self.parentObject = getParent(self.context)
        while (interfaces.IStudentCompetencyGradebook.providedBy(self.parentObject) or
               interfaces.ICompetencyGradebook.providedBy(self.parentObject)):
            self.parentObject = getParent(self.parentObject)

    @property
    def students(self):
        return sorted(self.context.students, key=lambda student: student.title)

    def competencyGradebooks(self):
        """return an iterator over the smaller gradebooks
        for each competency in this gradebook."""
        for comp in self.context.competencies.values():
            traverser = getMultiAdapter((self.context, self.request), IPublishTraverse)
            gb = traverser.publishTraverse(self.request, getName(comp))
            yield gb

    def studentGradebooks(self):
        """return the students gradebook for each student in the gradebook."""
        for student in self.students:
            traverser = getMultiAdapter((self.context, self.request), IPublishTraverse)
            gb = traverser.publishTraverse(self.request, '~'+getName(student))
            yield gb

    def studentGradebooksStatistics(self):
        """return the statistics for students gradebook for each student in the gradebook."""
        for stu_gb in self.studentGradebooks():
            yield interfaces.IStatistics(stu_gb)


class SectionGradebookGoView(BaseSectionGradebookView):
    """The 'go.html' view for the section competency gradebook."""

    def __init__(self, context, request):
        super(SectionGradebookGoView, self).__init__(context, request)

    def __call__(self):
        self.size = self._getSize()
        if self.size <= 100:
            url = absoluteURL(self.context, self.request) + '/@@index.html'
            self.request.response.redirect(url)
        return super(SectionGradebookGoView, self).__call__()

    def _getSize(self):
        hashDict = gradebook._buildHashDict(self.context.competencies)
        gradableComps = filter(ICompetencyReference.providedBy, hashDict.values())
        size = len(gradableComps)*len(list(self.students))
        return size


class SectionGradebookView(BaseSectionGradebookView):

    def expandHandlers(self):
        hashDict = gradebook._buildHashDict(self.context.competencies)
        results = "$(document).ready(function(){"
        results += "$('#%s').children('.gradebook-information').show()" \
            % hash(getRequirementKey(self.context.competencies))
        for compId, comp in hashDict.items():
            if ICompetencies.providedBy(comp):
                results += '''
                     $("#%s").children(".competency-title").click(function(){
                     $("#%s").children(".gradebook-information").toggle();});
                  ''' % (compId, compId)
            if not ICompetencies.providedBy(comp):
                results+='''
                     $("#%s-widgets-filldown").bind("change", function(){
                       $("#%s-widgets-filldown").parent().siblings().children(".evaluation-score")
                                                .val($("#%s-widgets-filldown").val());
                       $("#%s-widgets-savefilldown").show();
                     });
                     $("#%s-widgets-savefilldown").hide();
                     $("#%s-widgets-savefilldown").click(function(){
                       $("#%s-widgets-filldown").parent().siblings().children(".evaluation-score")
                                                .trigger("change");
                       $("#%s-widgets-savefilldown").hide();
                     });
                ''' % (compId, compId, compId, compId, compId, compId, compId, compId)
        results+="});"
        return results


class InlineSectionGradebookInstructorView(form.Form, BaseSectionGradebookView):

    template = ViewPageTemplateFile('templates/inlinesectiongradebookgrade.pt')
    filldown = zope.schema.TextLine(title=u"Fill Down")
    filldown.__name__ = 'filldown'
    savefilldown = zope.schema.TextLine(title=u"Save", default=u"Save")
    savefilldown.__name__ = 'savefilldown'
    fields = field.Fields(filldown, savefilldown)

    vertical = SessionProperty('vertical')

    def __init__(self, context, request):
        form.Form.__init__(self, context, request)
        self.prefix = str(hash(getRequirementKey(self.context.competencies)))

        self.isGroup = ICompetencies.providedBy(self.context.competencies)
        self.containsOnlyGroups = True
        for comp in self.context.competencies.values():
            if not interfaces.ICompetencies.providedBy(comp):
                self.containsOnlyGroups = False
                break

        self.style = ""
        if len(self.context.competencies.values()) != 0:
            self.style = "display: none;"

        vertical = self.request.get('vertical')
        if vertical == 'True':
            self.vertical = True
        elif vertical == 'False':
            self.vertical = False

    @property
    def statistics(self):
        stats = interfaces.IStatistics(self.context)
        result = {'average':stats.average,
                  'median':stats.median,
                  'standardDeviation':stats.standardDeviation,
                  'variance':stats.variance}
        for key, val in result.items():
            if val is None:
                result[key] = 'n/a'
            else:
                result[key] = '%0.2f' % val
        return result

    def updateWidgets(self):
        """See z3c.form.interfaces.IForm"""
        self.widgets = getMultiAdapter(
            (self, self.request, self.getContent()), IWidgets)
        self.widgets.ignoreContext = True
        self.widgets.update()
        self.widgets['filldown'].mode = 'input'
        self.widgets['filldown'].size = 3
        self.widgets['filldown'].addClass("fill-down")
        self.widgets['savefilldown'].mode = 'display'
        self.widgets['savefilldown'].size = 3
        self.widgets['savefilldown'].addClass("save-fill-down")
