#
# CanDo - Web based competency tracking system.
# Copyright (C) 2007 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

from datetime import datetime
from zope.keyreference.interfaces import IKeyReference
from zope.interface import Interface
import zope.schema
from zope.publisher.browser import BrowserView
from zope.component import getMultiAdapter, getUtility
from zope.publisher.interfaces import IPublishTraverse
from zope.traversing.api import getName, getParent, getParents
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
from zope.traversing.browser import absoluteURL
from zope.security.proxy import removeSecurityProxy

from z3c.form import form, field
from z3c.form.interfaces import IWidgets
from z3c.formjs import jsaction, jsevent, ajax
from reportlab.lib import units

from schooltool.app import relationships
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.report import ReportPDFView
from schooltool.course.interfaces import ISection
from schooltool.securitypolicy.crowds import TeachersCrowd
from schooltool.relationship import getRelatedObjects
from schooltool.requirement.requirement import getRequirementKey
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.app.browser.report import ReportPageTemplate

from cando.virginia.competency import getFlatCompetencies
from cando.virginia.interfaces import IUniversalScoreSystem
from cando.gradebook import interfaces
from cando.gradebook.browser import report
from cando.virginia.interfaces import ICompetencies
from cando.virginia.interfaces import ICompetencyReference


class IStudentGradebookFields(Interface):
    """Fields for a student gradebook."""

    score = zope.schema.TextLine(title=u"Score")
    date = zope.schema.TextLine(title=u"Date")


class StudentGradebookView(BrowserView):

    def __init__(self, context, request):
        super(StudentGradebookView, self).__init__(context, request)

        # find a parent object that isn't a gradebook for reference
        self.parentObject = getParent(self.context)
        while (interfaces.IStudentCompetencyGradebook.providedBy(self.parentObject) or
               interfaces.ICompetencyGradebook.providedBy(self.parentObject)):
            self.parentObject = getParent(self.parentObject)

        # find section object and build courses list (used in template)
        for parent in getParents(self.context):
            if ISection.providedBy(parent):
                section = parent
                break
        self.courses = getRelatedObjects(section, relationships.URICourse)
        self.schoolyear = ISchoolYear(section)
        self.instructors = ", ".join(["%s %s" % (instructor.first_name,
                                                 instructor.last_name,)
                                      for instructor in section.instructors])

        # provide easy access to the scoresystem
        self.scoresystem = ISchoolToolApplication(None)['scoresystem']

        # provide easy access to competencies
        self.competencies = removeSecurityProxy(context.__parent__).comprefDict
        self.groups = self.getSortedGroups()

        # calcluate summary stats
        self.numComps = self.numCompetent = self.totalPoints = 0
        for comp in getFlatCompetencies(self.competencies):
            ev = self.context.getEvaluation(comp)
            if ev is None or \
               self.scoresystem.getNumericalValue(ev.value) is None:
                continue
            self.numComps += 1
            if self.scoresystem.isPassingScore(ev.value):
                self.numCompetent += 1
            self.totalPoints += ev.value.value
        self.numCompsTimes3 = self.numComps * 3
        if self.numComps:
            self.percentCompetent = report.formatPercent(
                float(100 * self.numCompetent) / float(self.numComps))
            self.totalPercent = report.formatPercent(
                float(100 * self.totalPoints) / float(self.numCompsTimes3))
        else:
            self.percentCompetent = self.totalPercent = ''

    def groupsRecurse(self, obj, parent):
        if hasattr(obj, 'items'):
            for subitem in obj.items():
                self.groupsRecurse(subitem,obj)
        else:
            if not self.rowmap.has_key(obj[0]):
                self.rowmap[obj[0]] = []
            self.rowmap[obj[0]] += obj[1]

    def isCompetency(self, obj):
        return ICompetencyReference.providedBy(obj)

    def getSortedGroups(self):
        groups = []
        for key in self.competencies:
            if self.competencies[key]:
                firstComp = self.competencies[key][0]
                group = firstComp.__parent__
                groups.append([IKeyReference(group), firstComp.globalId,
                               key])
            else:
                groups.append([0, '', key])
        return [key for cid, fid, key in sorted(groups)]

    def getRows(self, group):
        comps = self.competencies.get(group, [])
        results = []
        for comp in comps:
            result = dict()
            result['competency'] = comp
            result['evaluation'] = self.context.getEvaluation(comp)
            results.append(result)
        return results


class StudentCompetencyRecordView(StudentGradebookView):

    def getSortedGroups(self):
        groups = []
        for key in self.competencies:
            if [comp for comp in self.competencies[key] if not comp.deprecated]:
                if self.competencies[key]:
                    firstComp = self.competencies[key][0]
                    group = firstComp.__parent__
                    groups.append([IKeyReference(group), firstComp.globalId,
                                   key])
                else:
                    groups.append([0, '', key])
        return [key for cid, fid, key in sorted(groups)]

    def getRows(self, group):
        comps = self.competencies.get(group, [])
        results = []
        for comp in comps:
            if not comp.deprecated:
                result = dict()
                result['competency'] = comp
                result['evaluation'] = self.context.getEvaluation(comp)
                results.append(result)
        return results


class InlineStudentGradebookStudentView(form.Form):
    template = ViewPageTemplateFile("templates/inlinestudentgradebook.pt")

    fields = field.Fields(IStudentGradebookFields)

    def __init__(self, context, request):
        super(InlineStudentGradebookStudentView, self).__init__(context, request)
        competencies = self.context.competencies
        self.prefix = "%s-%s" % (str(self.context.student.username),
                                 hash(getRequirementKey(competencies)))
        self.scoresystem = ISchoolToolApplication(None)['scoresystem']
        eval = self.context.getEvaluation(self.context.competencies, None)
        if eval is None or \
           self.scoresystem.getNumericalValue(eval.value) is None:
            self.score = u'-'
            self.date = u''
        else:
            self.score = int(eval.value)
            self.date = eval.time.strftime('%x')

        self.isGroup = ICompetencies.providedBy(competencies)


    def updateWidgets(self):
        '''See interfaces.IForm'''
        self.widgets = getMultiAdapter(
            (self, self.request, self.getContent()), IWidgets)
        self.widgets.ignoreContext = True
        self.widgets.update()
        self.widgets['score'].value = self.score
        self.widgets['score'].mode = 'display'
        self.widgets['score'].addClass('evaluation-score')
        self.widgets['date'].value = self.date
        self.widgets['date'].mode = 'display'
        self.widgets['date'].addClass('evaluation-date')

    def competencyGradebooks(self):
        """return an iterator over the smaller gradebooks
        for each competency in this gradebook."""
        for comp in self.context.competencies.values():
            traverser = getMultiAdapter((self.context, self.request), IPublishTraverse)
            gb = traverser.publishTraverse(self.request, getName(comp))
            yield gb

    @property
    def isTeacher(self):
        return TeachersCrowd(self.context).contains(self.request.principal)


class InlineStudentGradebookInstructorView(ajax.AJAXRequestHandler, InlineStudentGradebookStudentView):

    template = ViewPageTemplateFile("templates/inlinestudentgradebookgrade.pt")

    fields = field.Fields(IStudentGradebookFields)

    def updateWidgets(self):
        '''See interfaces.IForm'''
        super(InlineStudentGradebookInstructorView, self).updateWidgets()
        self.widgets['score'].mode = 'input'
        self.widgets['score'].size = 3

    @jsaction.handler(fields['score'], event=jsevent.CHANGE)
    def submitScoreHandler(self, event, selecter):
        '''the js handler that actually called the submitScore method via an http request'''
        scoreId = self.widgets['score'].id
        url = absoluteURL(self.context, self.request)
        return '''$.get("%s/@@inlinegrade/@@ajax/submitScore", {score: $("#%s").val(), id: "%s"},
                           function(data){
                              $.get("%s/@@inlinegrade/@@ajax/getDate", function(data){$("#%s").html(data)});
                              $("#message-notification").html(data);
                              $("#message-notification").highlightFade({color:'yellow',speed:2000});
                           }
                       );''' % (url, scoreId, self.prefix, url, self.widgets['date'].id)

    @ajax.handler
    def submitScore(self):
        '''Handle AJAX requests for updating a score'''
        try:
            score = self.request.get('score')
            comp = removeSecurityProxy(self.context.competencies)
            scoreObj = comp.scoresystem.fromUnicode(score)
            # XXX: This view is properly secure - but I still don't
            # like having removeSecurityProxy here.
#            context = removeSecurityProxy(self.context)
            eval = self.context.evaluate(comp, scoreObj)
            eval = removeSecurityProxy(eval)
        except zope.schema.ValidationError, e:
            return "There was an error: %r" % e
        return "Score %s saved for %s on %s for student %s." % (
            score,
            comp.competency.title,
            eval.time.strftime('%x %X'),
            self.context.student.title)

    @ajax.handler
    def getDate(self):
        '''Get the date for the evaluation.'''
        return self.date


class ValidatedCertificateView(ReportPDFView):

    template = ViewPageTemplateFile('templates/validatedcertificate.pt')

    def __init__(self, context, request):
        super(ValidatedCertificateView, self).__init__(context, request)
        ss = getUtility(IUniversalScoreSystem)
        # check 'Validated Resume' in cando.gradebook.README.txt
        self.byScoreAndGroup = {}
        for compref in getFlatCompetencies(self.context.comprefDict):
            if not compref.deprecated:
                evaluation = self.context.getEvaluation(compref)
                if evaluation and ss.isPassingScore(evaluation.value):
                    self.byScoreAndGroup.setdefault(evaluation.value, {})
                    group = compref.__parent__
                    self.byScoreAndGroup[evaluation.value].setdefault(group, [])
                    self.byScoreAndGroup[evaluation.value][group].append(compref)
        self.scoresOrder = sorted(self.byScoreAndGroup.keys(),
                                  key=lambda x:ss.getNumericalValue(x),
                                  reverse=True)
        section = self.context.__parent__.__parent__
        self.courses = ', '.join([course.title for course in section.courses])
        # XXX: repeated in the first page template
        doc_w, doc_h = self.pageSize
        self.signature = {
            'width': doc_w - self.leftMargin - self.rightMargin,
            'height': 0.75 * units.inch,
            'x': self.leftMargin,
            'y': self.bottomMargin + 0.6 * units.inch,
            'font-height': 0.2 * units.inch,
            }


class ValidatedCertificateFrontPageTemplate(ReportPageTemplate):

    template=ViewPageTemplateFile('templates/front_page_template.pt',
                                  content_type="text/xml")

    style_template=ViewPageTemplateFile('templates/style.pt',
                                        content_type="text/xml")

    footer_text = u"" # additional short footer text
    report_date = None # report generation date

    def __init__(self, *args, **kw):
        ReportPageTemplate.__init__(self, *args, **kw)
        self.report_date = datetime.today()
        self.frame = self.getFrame()

    @property
    def toplogo(self):
        doc_w, doc_h = self.parent.pageSize
        height = 1.0 * units.inch
        width = doc_w - self.parent.leftMargin - self.parent.rightMargin

        return {
            'height': height,
            'width': width,

            'x': self.parent.leftMargin,
            'y': doc_h - self.parent.topMargin - height
            }

    @property
    def header(self):
        if not self.parent.title:
            return {'height': 0}
        doc_w, doc_h = self.parent.pageSize
        font_height = 0.25 * units.inch
        header_height = font_height + 0.2 * units.inch
        return {
            'title': {
                'x': (doc_w/2.0),
                'y': (doc_h - self.parent.topMargin - font_height),
                },
            'height': header_height,
            }

    @property
    def footer(self):
        doc_w, doc_h = self.parent.pageSize

        footer_height = 0.6 * units.inch
        font_height = 10 # ~0.14 inch
        rule_y = (self.parent.bottomMargin + 0.55 * units.inch)

        return {
            'rule': '%d %d %d %d' % (
                self.parent.leftMargin, rule_y,
                doc_w - self.parent.rightMargin, rule_y),
            'logo': {
                'x': self.parent.leftMargin,
                'y': self.parent.bottomMargin,
                'height': 0.5 * units.inch,
                },
            'font_height': font_height,
            'center_place': {
                'x': self.parent.leftMargin,
                'y': (self.parent.bottomMargin + footer_height
                      - 0.2 * units.inch
                      - font_height),
                'width': (doc_w - self.parent.leftMargin - self.parent.rightMargin),
                'height': font_height + 0.1 * units.inch,
                },
            'right_str': {
                'x': (doc_w - self.parent.rightMargin),
                'y': (self.parent.bottomMargin + footer_height
                      - 0.1 * units.inch - font_height),
                },
            'right_str_2': {
                'x': (doc_w - self.parent.rightMargin),
                'y': (self.parent.bottomMargin + footer_height
                      - 0.15 * units.inch - font_height * 2),
                },
            'height': footer_height,
            }

    def getFrame(self):
        doc_w, doc_h = self.parent.pageSize
        fr = {
            'x': self.parent.leftMargin,
            'y': self.parent.bottomMargin + self.footer['height'],
            'width': doc_w - self.parent.leftMargin - self.parent.rightMargin,
            'height': (doc_h - self.parent.topMargin - self.parent.bottomMargin
                       - self.header['height'] - self.footer['height']),
            }
        height = (doc_h - self.parent.topMargin - self.parent.bottomMargin
                  - self.header['height'] - self.footer['height']
                  - self.signature['height'] - 0.25 * units.inch)
        fr.update({
                'height_with_header': height - self.toplogo['height'],
                'height': height,
                'y': (self.parent.bottomMargin + self.footer['height']
                      + self.signature['height'])
                })
        return fr

    @property
    def signature(self):
        doc_w, doc_h = self.parent.pageSize
        return {
            'width': doc_w - self.parent.leftMargin - self.parent.rightMargin,
            'height': 0.75 * units.inch,

            'x': self.parent.leftMargin,
            'y': self.parent.bottomMargin + self.footer['height'],
            
            'font-height': 0.2 * units.inch,
            }
