#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
"""Competency Gradebook Implementation

$Id$
"""
import persistent.dict

import zope.component
from zope import annotation
from zope.schema.fieldproperty import FieldProperty
from zope.security.proxy import removeSecurityProxy
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.location.location import LocationProxy
from zope.interface import implements
from zope.proxy import removeAllProxies
from zope.interface.declarations import alsoProvides

from schooltool.requirement.evaluation import Evaluation
from schooltool.requirement.interfaces import IEvaluations
from schooltool.app.interfaces import ISchoolToolApplication

from schooltool.course.interfaces import ISection
from schooltool.person.interfaces import IPerson
from schooltool.securitypolicy.crowds import Crowd

from cando.virginia.competency import buildCompetencyDict
from cando.virginia.interfaces import IScoreSystemRequirement
from cando.virginia.interfaces import IUniversalScoreSystem
from cando.skilldriver.interfaces import ISkillDriver

import interfaces

CURRENT_COMP_GROUP_KEY = 'cando.gradebook.currentcompgroup'


def getScoreSystem(competency):
    if IScoreSystemRequirement.providedBy(competency.__parent__):
        return competency.__parent__.scoresystem

    ss = zope.component.queryUtility(IUniversalScoreSystem)
    if ss is None:
        try:
            ss = ISchoolToolApplication(None).get('scoresystem', None)
        except ValueError:
            pass #failed to adapt

    return ss


# Common Helper Functions
def _getEvaluation(student, compref, default=None):
    """See interfaces.IGradebook"""
    evaluations = IEvaluations(student)
    return evaluations.get(compref.competency, default)


def _evaluate(student, compref, score, evaluator=None):
    """See interfaces.IGradebook"""
    evaluation = _getEvaluation(student, compref)
    competency = compref.competency
    if evaluation == None:
        evaluation = Evaluation(
            competency, getScoreSystem(competency),
            score, removeSecurityProxy(evaluator))
    elif evaluation.value != score:
        evaluation.value = score
        evaluation.evaluator = removeSecurityProxy(evaluator)
    evaluations = IEvaluations(student)
    evaluations.addEvaluation(evaluation)
    return evaluation


def stripCompare(str1, str2):
    return ' '.join(str1.split()) == ' '.join(str2.split())


class GradebookBase(object):

    def __init__(self, title, comprefDict):
        self.title = title
        self.comprefDict = comprefDict
        self.availableGlobalIds = set()
        for comps in self.comprefDict.values():
            self.availableGlobalIds.update([
                compref.globalId for compref in comps])

    def resetCurrentCompGroup(self, person):
        """Resets the current competency group to the first group
        alphabetically in the person's list."""
        person = removeSecurityProxy(person)
        if self.comprefDict:
            default = sorted(self.comprefDict.keys())[0]
        else:
            default = None
        self.setCurrentCompGroup(person, default)

    def getCurrentCompGroup(self, person):
        """Returns the current competency group of the person."""
        if not self.comprefDict:
            return None
        person = removeSecurityProxy(person)
        ann = annotation.interfaces.IAnnotations(person)
        if CURRENT_COMP_GROUP_KEY not in ann:
            ann[CURRENT_COMP_GROUP_KEY] = persistent.dict.PersistentDict()
        default = sorted(self.comprefDict.keys())[0]
        section_id = hash(self.__parent__)
        comp_group = ann[CURRENT_COMP_GROUP_KEY].get(section_id, default)
        for key in self.comprefDict.keys():
            if stripCompare(comp_group, key):
                return comp_group
        return default

    def setCurrentCompGroup(self, person, compgroup):
        """Sets the current competency group of the person."""
        person = removeSecurityProxy(person)
        compgroup = removeSecurityProxy(compgroup)
        ann = annotation.interfaces.IAnnotations(person)
        if CURRENT_COMP_GROUP_KEY not in ann:
            ann[CURRENT_COMP_GROUP_KEY] = persistent.dict.PersistentDict()
        section_id = hash(self.__parent__)
        ann[CURRENT_COMP_GROUP_KEY][section_id] = compgroup

    def _checkCompetency(self, compref):
        if compref.globalId not in self.availableGlobalIds:
            raise ValueError(
                '%r is not part of this gradebook.' % compref.competency.title)
        # Remove security proxy, so that the object can be referenced and
        # adapters are not proxied. Note that the gradebook itself has
        # sufficiently tight security.
        return removeSecurityProxy(compref)


# Gradebook Implementations

class CompetencyGradebook(GradebookBase):
    implements(interfaces.ICompetencyGradebook,
               interfaces.IGradebookUtilities,
               interfaces.IHaveGradebookInstructors,
               IAttributeAnnotatable)

    def __init__(self, title, students, comprefDict):
        super(CompetencyGradebook, self).__init__(title, comprefDict)
        self.students = students

    def _checkStudent(self, student):
        if student not in self.students:
            raise ValueError(
                'Student %r is not in this gradebook.' % student.title)
        # Remove security proxy, so that the object can be referenced and
        # adapters are not proxied. Note that the gradebook itself has
        # sufficiently tight security.
        return removeSecurityProxy(student)

    def getEvaluation(self, student, compref, default=None):
        """See interfaces.IGradebook"""
        student = self._checkStudent(student)
        compref = self._checkCompetency(compref)
        return _getEvaluation(student, compref, default)

    def evaluate(self, student, compref, score, evaluator=None):
        """See interfaces.IGradebook"""
        student = self._checkStudent(student)
        compref = self._checkCompetency(compref)
        score = removeAllProxies(score) # remove proxy for storage
        return removeAllProxies(_evaluate(student, compref, score, evaluator))


class StudentCompetencyGradebook(GradebookBase):
    implements(interfaces.IStudentCompetencyGradebook,
                              interfaces.IGradebookUtilities,
                              interfaces.IHaveGradebookInstructors,
                              IAttributeAnnotatable)

    title = FieldProperty(interfaces.IStudentCompetencyGradebook['title'])
    student = FieldProperty(interfaces.IStudentCompetencyGradebook['student'])

    def __init__(self, title, student, comprefDict):
        super(StudentCompetencyGradebook, self).__init__(title, comprefDict)
        self.student = student

    def getEvaluation(self, compref, default=None):
        """See interfaces.IStudentCompetencyGradebook"""
        compref = self._checkCompetency(compref)
        return _getEvaluation(self.student, compref, default)

    def evaluate(self, compref, score, evaluator=None):
        """See interfaces.IGradebook"""
        compref = self._checkCompetency(compref)
        score = removeSecurityProxy(score) # remove proxy for storage
        return _evaluate(self.student, compref, score, evaluator)


# Adapters
def sectionGradebook(obj):
    """adapter from a section to a competency gradebook."""
    section = obj
    while not ISection.implementedBy(section.__class__):
        section = section.__parent__

    # for skill drivers, we'll flatten out the competency dict
    comps = buildCompetencyDict(obj)
    if ISkillDriver.providedBy(obj):
        newComps = []
        for compList in comps.values():
            newComps.extend(compList)
        comps = {obj.title: newComps}

    gb = CompetencyGradebook(unicode(obj.title),
                             section.members,
                             comps)
    interfaces.IGradebookInstructors(gb).instructors = list(section.instructors)
    gb = LocationProxy(gb, obj, 'gradeCompetencies')
    gb.__setattr__('__parent__', gb.__parent__) # XXX THIS IS REALLY BAD XXX
    if ISection.implementedBy(obj.__class__):
        alsoProvides(gb, interfaces.IHaveScoreReports)
    return gb


def studentGradebook(student, gradebook):
    """adapter from student, comp gradebook to student competency gradebook."""
    if student not in gradebook.students:
        raise ValueError("%r is not a student in %s" % (student, gradebook.title))
    title = u"%s's Gradebook" % getattr(student, 'title', 'Student')
    gb = StudentCompetencyGradebook(title, student, gradebook.comprefDict)
    gb.__parent__ = gradebook
    if ISection.implementedBy(gb.__parent__.__parent__.__class__):
        alsoProvides(gb, interfaces.IHaveStudentScoreReports)
    interfaces.IGradebookInstructors(gb).instructors = interfaces.IGradebookInstructors(gradebook).instructors
    return gb


class GradebookInstructors(object):
    """See interfaces.IGradebookInstructors"""
    implements(interfaces.IGradebookInstructors)
    zope.component.adapts(interfaces.IHaveGradebookInstructors)

    instructors = FieldProperty(interfaces.IGradebookInstructors['instructors'])

getGradebookInstructors = zope.annotation.factory(GradebookInstructors, 'instructors')


# Security Crowds
class GradebookInstructorsCrowd(Crowd):
    """Crowd of gradebook graders."""
    zope.component.adapts(interfaces.IHaveGradebookInstructors)
    def contains(self, principal):
        instructor = IPerson(principal, None)
        return instructor in interfaces.IGradebookInstructors(self.context).instructors


class GradebookStudentCrowd(Crowd):
    """Crowd containing just the student of a student gradebook."""
    zope.component.adapts(interfaces.IStudentCompetencyGradebook)
    def contains(self, principal):
        student = IPerson(principal, None)
        return student == self.context.student


class GradebookStudentsCrowd(Crowd):
    """Crowd containing just the student of a student gradebook."""
    zope.component.adapts(interfaces.ICompetencyGradebook)
    def contains(self, principal):
        student = IPerson(principal, None)
        return student in self.context.students
