#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import math

from zope.component import adapter
import zope.interface
from zope.schema.fieldproperty import FieldProperty
from schooltool.requirement.scoresystem import UNSCORED

import interfaces, gradebook


class NumberStatistics(object):
    """This implementation only deals with numbers."""
    zope.interface.implements(interfaces.IStatistics)
    
    values = FieldProperty(interfaces.IStatistics['values'])
    average = FieldProperty(interfaces.IStatistics['average'])
    median = FieldProperty(interfaces.IStatistics['median'])
    standardDeviation = FieldProperty(interfaces.IStatistics['standardDeviation'])
    variance = FieldProperty(interfaces.IStatistics['variance'])

    def __init__(self, values):
        self.values = tuple([float(v) for v in values])
        self.average = self._calculateAverage()
        self.median = self._calculateMedian()
        self.variance = self._calculateVariance()
        self.standardDeviation = self._calculateStandardDeviation()

    def _calculateAverage(self):
        if len(self.values) == 0:
            return None
        return sum(self.values)/len(self.values)

    def _calculateMedian(self):
        """See interfaces.IStatistics"""
        size = len(self.values)
        if size == 0:
            return None

        sortedValues = sorted(self.values)
        return sortedValues[size/2]

    def _calculateStandardDeviation(self):
        """See interfaces.IStatistics"""
        variance = self.variance

        if variance is None:
            return None

        return float(math.sqrt(variance))

    def _calculateVariance(self):
        """See interfaces.IStatistics"""
        avg = self.average

        if avg is None:
            return None

        help = [(item - avg)**2 for item in self.values]

        if len(self.values) == 1:
            return None

        return sum(help)/(len(self.values)-1)

@adapter(interfaces.IStudentCompetencyGradebook)
@zope.interface.implementer(interfaces.IStatistics)
def studentGradebookStatistics(gb):
    """Adapter from student gradebook to statistics."""
    comps = gradebook._getFlatCompetencies(gb.competencies)
    evals = filter(lambda e: e is not None and e.value is not UNSCORED,
                   [gb.getEvaluation(comp) for comp in comps])
    scores = [e.value for e in evals]
    return NumberStatistics(scores)


@adapter(interfaces.ICompetencyGradebook)
@zope.interface.implementer(interfaces.IStatistics)
def sectionGradebookStatistics(gb):
    """Adapter from student gradebook to statistics."""
    comps = gradebook._getFlatCompetencies(gb.competencies)
    evals = []
    for student in gb.students:
        evals += filter(lambda e: e is not None and e.value is not UNSCORED,
                        [gb.getEvaluation(student, comp) for comp in comps])
    scores = [e.value for e in evals]
    return NumberStatistics(scores)
