#
# CanDo - Web based competency tracking system.
# Copyright (C) 2010 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Student IEP views

$Id$
"""

import urllib
from persistent.dict import PersistentDict
from persistent.list import PersistentList
from zope.publisher.browser import BrowserView
from zope.traversing.browser import absoluteURL
from zope.i18nmessageid import MessageFactory
from zope.app.renderer.plaintext import PlainTextToHTMLRenderer

from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.term.interfaces import ITerm
from schooltool.course.interfaces import ILearner, ISectionContainer
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.schoolyear.interfaces import ISchoolYearContainer

from cando.virginia.competency import buildCompetencyDict
from cando.virginia.competency import getAdditionalCompetencies
from cando.iep.interfaces import IStudentIEP

_ = MessageFactory('cando')


class StudentRedirect(object):

    nextURL = None

    def redirectToPersonView(self):
        self.nextURL = 'index.html'
        self.redirect()

    def redirectToStudentIEPView(self):
        self.nextURL = 'iep.html'
        self.redirect()

    def redirectToCreateStudentIEPView(self):
        self.nextURL = 'iep_create.html'
        self.redirect()

    def redirectToCommentStudentIEPView(self):
        self.nextURL = 'iep_comment.html'
        self.redirect()

    def redirect(self):
        url = "%s/%s" % (absoluteURL(self.context, self.request),
                         self.nextURL or '')
        self.request.response.redirect(url)


class StudentIEPView(BrowserView, StudentRedirect):

    def __init__(self, context, request):
        super(StudentIEPView, self).__init__(context, request)
        self.student_iep = IStudentIEP(self.context)
        if not self.student_iep.hasIEP:
            return self.redirectToCreateStudentIEPView()

    def update(self):
        self.comment = PlainTextToHTMLRenderer(self.student_iep.comment or '',
                                               self.request).render()
        sections = []
        for section in ILearner(self.context).sections():
            info = {}
            info['school_year'] = ISchoolYear(section)
            info['school_year_title'] = ISchoolYear(section).title
            info['term'] = ITerm(section)
            info['term_title'] = ITerm(section).title
            info['courses'] = list(section.courses)
            info['courses_titles'] = ', '.join([course.title
                                                for course in section.courses])
            info['section'] = section
            info['section_title'] = section.title
            url_params = urllib.urlencode([('school_year',
                                            ISchoolYear(section).__name__),
                                           ('term', ITerm(section).__name__),
                                           ('section', section.__name__)]
                                          )
            info['section_url'] = absoluteURL(self.context, self.request) + \
                                  '/iep_competencies.html?%s' % (url_params,)
            sections.append(info)
        self.sections = sorted(sections,
                               key=lambda x:(x['school_year'].first,
                                             x['term'].first,
                                             x['courses_titles'],
                                             x['section_title']))


class CreateStudentIEPView(BrowserView, StudentRedirect):

    def update(self):
        if 'yes' in self.request:
            student_iep = IStudentIEP(self.context)
            if not student_iep.hasIEP:
                student_iep.hasIEP = True
                return self.redirectToCommentStudentIEPView()
            else:
                return self.redirectToStudentIEPView()
        elif 'no' in self.request:
            return self.redirectToPersonView()


class CommentStudentIEPView(BrowserView, StudentRedirect):

    def __init__(self, context, request):
        super(CommentStudentIEPView, self).__init__(context, request)
        self.student_iep = IStudentIEP(self.context)
        if not self.student_iep.hasIEP:
            return self.redirectToCreateStudentIEPView()

    def update(self):
        self.comment = self.student_iep.comment or ''
        if 'save' in self.request:
            if 'comment' in self.request:
                self.student_iep.comment = self.request['comment']
            return self.redirectToStudentIEPView()


# XXX: Mostly taken from cando.virginia.browser.virginia.ViewCompetencies
class CompetenciesStudentIEPView(BrowserView, StudentRedirect):

    def __init__(self, context, request):
        super(CompetenciesStudentIEPView, self).__init__(context, request)
        self.section = None
        self.school_year_title = ''
        self.student_iep = IStudentIEP(self.context)
        school_year_name = self.request.get('school_year')
        term_name = self.request.get('term')
        section_name = self.request.get('section')
        app = ISchoolToolApplication(None)
        try:
            schoolyears = ISchoolYearContainer(app)
            self.school_year = schoolyears[school_year_name]
            self.school_year_title = self.school_year.title
            self.term = self.school_year[term_name]
            sections = ISectionContainer(self.term)
            self.section = sections[section_name]
        except (KeyError,):
            return self.redirectToStudentIEPView()

    def update(self):
        self.additionalComps = {}
        self.comps = {}
        changed = False
        if self.section is not None:
            school_year_name = self.school_year.__name__
            section_name = self.section.__name__
            competencies = self.student_iep.competencies
            sy_comps = competencies.setdefault(school_year_name,
                                               PersistentDict())
            if 'save' in self.request:
                if 'iep_comps' in self.request:
                    iep_comps = self.request['iep_comps']
                    if not isinstance(iep_comps, list):
                        iep_comps = [iep_comps,]
                    section_comps = sy_comps.setdefault(section_name,
                                                        PersistentList())
                    if sorted(section_comps) != sorted(iep_comps):
                        sy_comps[section_name] = PersistentList(iep_comps)
                        changed = True
                elif sy_comps.get(section_name):
                    # no comps were checked. erase them all
                    sy_comps[section_name] = PersistentList()
                    changed = True
                if changed:
                    self.message = _("The changes were saved")
                else:
                    self.message = _("No changes were made")
            elif 'cancel' in self.request:
                return self.redirectToStudentIEPView()
            self.comps = buildCompetencyDict(self.section)
            self.getSectionAdditionalComps()
            self.getCoursesAdditionalComps()

    def groups(self):
        core_groups = []
        additional_groups = []
        for group_title in self.comps:
            if self.comps[group_title]:
                firstComp = self.comps[group_title][0]
                if self.isAdditionalCompetency(firstComp):
                    additional_groups.append(group_title)
                else:
                    core_groups.append(group_title)
        return sorted(core_groups, key=self.sortKey) + \
               sorted(additional_groups, key=self.sortKey)

    def sortKey(self, group_title):
        ids = sorted([compref.id for compref in self.comps[group_title]])
        if ids:
            return ids[0]
        else:
            return ''

    def getSectionAdditionalComps(self):
        section = self.getSection()
        for comp in getAdditionalCompetencies(section):
            self.additionalComps[comp] = None

    def getCoursesAdditionalComps(self):
        courses = self.getCourses()
        for course in courses:
            for comp in getAdditionalCompetencies(course):
                self.additionalComps[comp] = None

    def getSection(self):
        return self.section

    def getCourses(self):
        return list(self.getSection().courses)
        
    def isAdditionalCompetency(self, comp):
        return comp in self.additionalComps

    def isIEPCompetency(self, comp):
        return self.student_iep.isIEPCompetency(self.section, comp)

    def getCompIdClass(self, comp):
        if self.isAdditionalCompetency(comp):
            return 'additional-comp'
        if self.isIEPCompetency(comp):
            return 'iep-comp'
        return '%s-comp' % comp.optionalFlag
