#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
from datetime import datetime

import persistent.dict
from zope.component import adapts, queryMultiAdapter, queryUtility
from zope.interface import implements, directlyProvides
from zope.annotation.interfaces import IAttributeAnnotatable, IAnnotations
from zope.container import btree
from zope.container.interfaces import INameChooser
from zope.publisher.interfaces import NotFound, IPublishTraverse
from zope.location import location
from zope.security import proxy
from zope.keyreference.interfaces import IKeyReference
from zope.traversing.api import getParent

from schooltool.course.interfaces import ISectionContainer
from schooltool.person.interfaces import IPerson
from schooltool.relationship import RelationshipProperty
from schooltool.securitypolicy.crowds import Crowd
from schooltool.app.app import getSchoolToolApplication
from schooltool.relationship import URIObject
from schooltool.term.interfaces import IDateManager
from schooltool.traverser.traverser import NameTraverserPlugin

import interfaces
from interfaces import IMessageBox, IMessage, IQueueItem

QUEUE_KEY = "cando.queue"
QUEUE_ITEM_KEY = "cando.queueItem"
MESSAGEBOX_KEY = "cando.message"

URIMessageBoxManagership = URIObject('http://schooltool.org/ns/messageboxManagership',
                          'Managership', 'The managership relationship.')
URIMessageBoxManager = URIObject('http://schooltool.org/ns/managership/messageboxManager',
                              'MessageBox Manager', 'A messagebox manager role.')
URIMessageBoxMembership = URIObject('http://schooltool.org/ns/messageboxMembership',
                     'MessageBox Membership', 'The membership to a messagebox')
URIMessageBox = URIObject('http://schooltool.org/ns/membership/messagebox',
                     'MessageBox', 'A role of a containing messagebox.')
URIMessageBoxMember = URIObject('http://schooltool.org/ns/membership/mesageboxMember',
                             'MessageBox Member', 'A messagebox member role.')


class MessageBox(btree.BTreeContainer):
    """MessageBox"""

    implements(interfaces.IMessageBox,
               IAttributeAnnotatable)

    members = RelationshipProperty(URIMessageBoxMembership,
                                   URIMessageBox,
                                   URIMessageBoxMember)

    managers = RelationshipProperty(URIMessageBoxManagership,
                                    URIMessageBox,
                                    URIMessageBoxMember)

    def __init__(self, title, section, person=None):
        super(MessageBox, self).__init__()
        self.title = title
        self.description = title
        self.person = person
        self.section = section

    def __repr__(self):
        return '<%s "%s">' % (self.__class__.__name__, self.title)


class MessageBoxMemberCrowd(Crowd):
    """Crowd of members of a messagebox."""
    adapts(IMessageBox)
    def contains(self, principal):
        if IQueueItem.providedBy(self.context):
            messageBox = self.context.context.__parent__
        else:
            messageBox = self.context
        if IPerson(principal, None) in messageBox.members:
            return True


class MessageBoxManagerCrowd(Crowd):
    """Crowd of managers of a messagebox."""
    adapts(IMessageBox)
    def contains(self, principal):
        return IPerson(principal, None) in self.context.managers


class MessageMemberCrowd(Crowd):
    """MessageBox members for a message"""
    adapts(IMessage)
    def contains(self, principal):
        messagebox = getParent(self.context)
        while not IMessageBox.providedBy(messagebox):
            messagebox = getParent(messagebox)
        return IPerson(principal, None) in messagebox.members


class MessageManagerCrowd(Crowd):
    """MessageBox managers for a message"""
    adapts(IMessage)
    def contains(self, principal):
        messagebox = getParent(self.context)
        while not IMessageBox.providedBy(messagebox):
            messagebox = getParent(messagebox)
        return IPerson(principal, None) in messagebox.managers


class Message(btree.BTreeContainer):
    """Message"""

    implements(interfaces.IMessage,
               interfaces.IMessageContained,
               IAttributeAnnotatable)

    def __init__(self, title, body, author=None,
                 created=None, modified=None):
        super(Message, self).__init__()
        self.title = title
        self.body = body
        self.author = author
        self.created = created or datetime.utcnow()
        self.modified = modified or datetime.utcnow()


class MessageBoxContainer(btree.BTreeContainer):
    """Container of MessageBoxes"""

    implements(interfaces.IMessageBoxContainer)

    def __init__(self):
        super(MessageBoxContainer, self).__init__()
        self._index = persistent.dict.PersistentDict()

    def _getHashKey(self, person, section):
        if person is not None:
            personHash = hash(IKeyReference(person, None))
        else:
            personHash = None
        if section is not None:
            sectionHash = hash(IKeyReference(section, None))
        else:
            sectionHash = None
        return (personHash, sectionHash)

    def add(self, messageBox):
        chooser = INameChooser(self)
        name = chooser.chooseName('',messageBox)
        self[name] = messageBox
        hashKey = self._getHashKey(messageBox.person, messageBox.section)
        self._index[hashKey] = messageBox

    def getMessageBox(self, person, section):
        return self._index.get(self._getHashKey(person, section), None)

    def remove(self, person, section):
        del self[self.getMessageBox(person, section).__name__]
        hashKey = self._getHashKey(person, section)
        del self._index[hashKey]


def getMessageBox(context):
    """Adapt an ``IHaveMessageBox`` object to ``IMessageBox``."""
    annotations = IAnnotations(context)
    messageBox = annotations[MESSAGEBOX_KEY]
    return messageBox

# Convention to make adapter introspectable
getMessageBox.factory = MessageBox


class PersonMessageBoxesProxy(object):
    """Proxy to a list of a person's messageboxes."""
    def __init__(self, person):
        self.person = person


class SectionMessageBoxesProxy(object):
    """Proxy to a list of a person's messageboxes."""
    def __init__(self, section):
        self.section = section

def addMessageBoxContainerToApplication(event):
    event.object['cando.message.message_box_container'] = MessageBoxContainer()

class SectionMessageBoxTraverserPlugin(NameTraverserPlugin):
    """Allows traversal to a forum from a section object.
    
    i.e. if you go to localhost/sections/somesec/forum this traverser
    will get called instead of standard traverser.  If the blog
    does not exist, it gets created.
    """
    traversalName = "forum"
    def _traverse(self, request, name):
        context = proxy.removeSecurityProxy(self.context)
        messageBoxes = getSchoolToolApplication()['cando.message.message_box_container']
        messageBox = messageBoxes.getMessageBox(None, context)
        if messageBox is None:
            messageBox = MessageBox(context.title+u"'s Forum", context)
            messageBox.members = context.members
            messageBox.managers = context.instructors
            messageBoxes.add(messageBox)
        messageBox = location.LocationProxy(messageBox, self.context,
                                         self.traversalName)
        directlyProvides(messageBox, interfaces.IForum)
        return messageBox

class PersonMessageBoxesTraverserPlugin(NameTraverserPlugin):
    """Allows traversal to a fake messagebox container for student-section messageboxes.
    
    i.e. if you go to localhost/persons/someone/messageboxes this traverser
    will get called instead of standard traverser.  It creates the fake messageboxes
    container on the fly.
    """
    traversalName = "messageboxes"
    def _traverse(self, request, name):
        proxyMessageBox = PersonMessageBoxesProxy(self.context)
        return location.LocationProxy(proxyMessageBox, self.context,
                                         self.traversalName)

class SectionMessageBoxesTraverserPlugin(NameTraverserPlugin):
    """Allows traversal to a fake messagebox container for student-section messageboxes.
    
    i.e. if you go to localhost/sections/somesec/messageboxes this traverser
    will get called instead of standard traverser.  It creates the fake messageboxes
    container on the fly.
    """
    traversalName = "messageboxes"
    def _traverse(self, request, name):
        proxyMessageBox = SectionMessageBoxesProxy(self.context)
        return location.LocationProxy(proxyMessageBox, self.context,
                                         self.traversalName)

class PersonMessageBoxesTraverser(object):
    """Allows traversal to a person-section messagebox from the person messageboxes
    proxy.
    
    i.e. if you go to localhost/persons/someone/messageboxes/somesection this
    traverser will get called instead of standard traverser.  It also wraps the
    messagebox in a location proxy so the url looks like what is above.
    """
    implements(IPublishTraverse)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def publishTraverse(self, request, name):
        app = getSchoolToolApplication()
        term = queryUtility(IDateManager).current_term
        sections = ISectionContainer(term)
        section = proxy.removeSecurityProxy(sections.get(name, None))
        person = proxy.removeSecurityProxy(self.context.person)

        if section is not None:
            messageBoxes = app['cando.message.message_box_container']
            messageBox = messageBoxes.getMessageBox(person, section)
            if messageBox is None:
                title = u"%s's MessageBox for Section %s" % (
                    person.title, section.title)
                messageBox = MessageBox(title, section, person)
                messageBox.members.add(person)
                messageBoxes.add(messageBox)
                messageBox.managers = section.instructors

            messageBox = location.LocationProxy(messageBox, self.context,
                                             section.__name__)
            directlyProvides(messageBox, interfaces.ISectionMessageBox)
            return messageBox

        view = queryMultiAdapter((self.context, request), name=name)
        if view is not None:
            return view
        raise NotFound(self.context, name, request)

class SectionMessageBoxesTraverser(object):
    """Allows traversal to a person-section messagebox from the person messageboxes
    proxy.
    
    i.e. if you go to localhost/sections/somesec/messageboxes/someone this
    traverser will get called instead of standard traverser.  It also wraps the
    messagebox in a location proxy so the url looks like what is above.
    """
    implements(IPublishTraverse)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def publishTraverse(self, request, name):
        app = getSchoolToolApplication()
        section = proxy.removeSecurityProxy(self.context.section)
        person = proxy.removeSecurityProxy(app['persons'].get(name, None))
        if person is not None:
            messageBoxes = app['cando.message.message_box_container']
            messageBox = messageBoxes.getMessageBox(person, section)
            if messageBox is None:
                title =u"%s's MessageBox for Section %s" % (
                    person.title, section.title)
                messageBox = MessageBox(title, section, person)
                messageBox.members.add(person)
                messageBoxes.add(messageBox)
            messageBox.managers = section.instructors
            messageBox = location.LocationProxy(messageBox, self.context,
                person.__name__)
            messageBox.section = section
            messageBox.person = person
            directlyProvides(messageBox, interfaces.ISectionMessageBox)
            return messageBox

        view = queryMultiAdapter((self.context, request), name=name)
        if view is not None:
            return view
        raise NotFound(self.context, name, request)


def locateMessageBox(obj):
    if IMessageBox.providedBy(obj):
        return obj
    return locateMessageBox(getParent(obj))
