#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
from zope.interface import implements
from zope.component import queryUtility
from z3c.layer.pagelet import IPageletBrowserLayer
from z3c.form.interfaces import IFormLayer
from jquery.layer import IJQueryJavaScriptBrowserLayer
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser.absoluteurl import AbsoluteURL
from zope.security.proxy import removeSecurityProxy
from zope.app.basicskin.standardmacros import StandardMacros

from schooltool.skin.skin import ISchoolToolLayer
from schooltool.person.interfaces import IPerson
from schooltool.course.interfaces import ISectionContainer
from schooltool.relationship import getRelatedObjects
from schooltool.app import relationships
from schooltool.securitypolicy.crowds import TeachersCrowd, StudentsCrowd
from schooltool.securitypolicy.crowds import ManagersCrowd
from schooltool.basicperson.browser.skin import IBasicPersonLayer
from schooltool.term.interfaces import IDateManager


class IFormSupportLayer(IFormLayer, IJQueryJavaScriptBrowserLayer):
    """Layer for support of z3c.form and z3c.formjs."""


class INewCanDoLayer(IFormSupportLayer, IPageletBrowserLayer,
                     IBasicPersonLayer, ISchoolToolLayer):
    """New CanDo layer."""


class INewCanDoSkin(INewCanDoLayer):
    """The same CanDo layer, BBB naming."""


class CanDoMacros(StandardMacros):
    macro_pages = ('newskin_macros',)


class AJAXAbsoluteURL(AbsoluteURL):
    """A special absolute URL getter that works with IAJAXHandler objects.

    This is for formjs support.  Eventually this should probably get put into
    formjs itself.  But as we are only at version 0.2 we have to implement it
    ourselves."""

    implements(IAbsoluteURL)

    def _getContextName(self, context):
        return removeSecurityProxy(context).func.__name__


class PersonDistinguisher(object):
    """Base class for any view that needs person distinguishing logic."""

    @property
    def person(self):
        return IPerson(self.request.principal)

    @property
    def isTeacher(self):
        return TeachersCrowd(self.context).contains(self.request.principal)

    @property
    def isStudent(self):
        return StudentsCrowd(self.context).contains(self.request.principal)
        
    @property
    def isManager(self):
        return ManagersCrowd(self.context).contains(self.request.principal)


class ChooserBase(PersonDistinguisher):
    """Base class for course and section choosers."""

    def _getSections(self):
        #sections = getRelatedObjects(IPerson(self.request.principal),
        #                             relationships.URISection)
        person = IPerson(self.request.principal); 
        term = queryUtility(IDateManager).current_term
        for key, section in ISectionContainer(term).items():
            teachers = [removeSecurityProxy(t) for t in section.instructors]
            students = [removeSecurityProxy(t) for t in section.members]

            if self.isTeacher and person in teachers:
                yield section
            if self.isStudent and person in students:
                yield section

class CourseChooser(ChooserBase):
    """provides a method to get courses."""

    def courses(self):
        """Returns a list of dictionaries that have the following keys:
              course - the course itself
              teachers - all teachers associated with the course
        """
        results = []
        searchFilter = self.request.get('filter')
        
        for section in self._getSections():
            courses = getRelatedObjects(section, relationships.URICourse)
            teachers = [removeSecurityProxy(t) for t in section.instructors]
            
            for course in courses:
                courseIsRelevant = False
                if searchFilter is None or "":
                    courseIsRelevant = True
                else:
                    searchFilter = searchFilter.lower()
                    if searchFilter in course.title.lower():
                        courseIsRelevant = True
                if courseIsRelevant:
                    self._includeCourse(results, course, teachers)
        
        return results
        
    def _includeCourse(self, results, course, teachers):
        for result in results:
            if result['course'] is course:
                break
        else:
            result = None
        if result:
            newTeachers = [teacher for teacher in teachers if teacher not in result['teachers']]
            result['teachers'].extend(newTeachers)
        else:
            results.append({'course': course, 'teachers': teachers})


class SectionChooser(ChooserBase):
    """provides a method to get sections."""

    def sections(self):
        """Returns a list of dictionaries that have the following keys:
              section - the section itself.
              courses - the courses associated with the section
              teachers - the teachers associated with the section
        """
        results = []
        searchFilter = self.request.get('filter')
        
        for section in self._getSections():
            courses = getRelatedObjects(section, relationships.URICourse)
            teachers = [removeSecurityProxy(t) for t in section.instructors]
            
            if searchFilter is None or "":
                results.append({'section': section,
                                'courses': courses, 
                                'teachers': teachers})
            else:
                searchFilter = searchFilter.lower()
                if searchFilter in [course.title.lower() for course in courses] or searchFilter in section.title.lower():
                    results.append({'section': section,
                                    'courses': courses, 
                                    'teachers': teachers})
        return results
