#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import csv
from cStringIO import StringIO
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent
from zope.component import getUtility
from zope.security.proxy import removeSecurityProxy
from zope.container.interfaces import INameChooser
from zope.i18nmessageid import MessageFactory
from zope.i18n import translate

from schooltool.common import parse_date
from schooltool.term.interfaces import IDateManager
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.csvimport import BaseCSVImportView
from schooltool.app.browser.csvimport import ImportErrorCollection
from schooltool.group.interfaces import IGroupContainer
from schooltool.basicperson.demographics import PersonDemographicsData
from schooltool.basicperson.person import BasicPerson
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.section import Section

from cando.virginia.competency import Competencies
from cando.virginia.competency import Competency, CompetencyReference
from cando.virginia.competency import getAppGlobalCompetencies


_ethnicity_codes = {
    'a': u'Asian',
    'b': u'Black',
    'h': u'Hispanic',
    'i': u'American Indian',
    'u': u'Unspecified',
    'w': u'White',
    }


_ = MessageFactory('cando')


class InvalidCSVError(Exception):
    pass

class CompetencyCSVImporter(object):
    """A Person CSV importer."""

    def __init__(self, container, charset):
        self.container = container
        self.charset = charset
        self.errors = ImportErrorCollection()
        self.global_comps = getAppGlobalCompetencies(
            ISchoolToolApplication(None))

    def parseCSVRows(self, rows):
        """Parse rows (a list of strings) in CSV format.

        Returns a list of rows as lists.  Trailing empty cells are discarded.

        rows must be in the encoding specified during construction of
        CompetencyCSVImportView; the returned values are in unicode.

        If the provided data is invalid, self.errors.generic will be updated
        and InvalidCSVError will be returned.
        """
        result = []
        reader = csv.reader(rows)
        line = 0
        try:
            while True:
                line += 1
                values = [v.strip() for v in reader.next()]
                if self.charset:
                    values = [unicode(v, self.charset) for v in values]
                # Remove trailing empty cells.
                while values and not values[-1].strip():
                    del values[-1]
                result.append(values)
        except StopIteration:
            return result
        except csv.Error:
            error_msg = _("Error in timetable CSV data, line ${line_no}",
                          mapping={'line_no': line})
            self.errors.generic.append(error_msg)
            raise InvalidCSVError()
        except UnicodeError:
            error_msg = _("Conversion to unicode failed in line ${line_no}",
                          mapping={'line_no': line})
            self.errors.generic.append(error_msg)
            raise InvalidCSVError()

    def _getIndentation(self, row):
        i = 0
        while not row[i]:
            i += 1
        return i

    def importCompetencies(self, rows, container=None, indent=0, line=0):
        if container == None:
            container = self.container

        lastContainer = container
        while line < len(rows):
            row = rows[line]
            indentation = self._getIndentation(row)
            if indentation > indent+1:
                self.errors.generic.append(_("Too many indents on line ${line}",
                                     mapping={'line':line}))
                raise InvalidCSVError()
            if indentation < indent: #end of block
                break
            if indentation == indent+1: #new block
                line = self.importCompetencies(rows, lastContainer, indent+1, line)
                continue

            objectData = row[indent:]
            if len(objectData) < 2:
                self.errors.generic.append(_("Not enough data on line ${line}",
                                     mapping={'line':line}))
                raise InvalidCSVError()
            objectType = objectData[0]
            if objectType == u'group':
                lastContainer = self._makeCompetencyGroup(container,
                                                          objectData[1:],
                                                          line)
            elif objectType == u'single':
                self._makeCompetency(container,
                                             objectData[1:],
                                             line)
            else:
                self.errors.generic.append(_("unrecognized type ${type} on line ${line}",
                                mapping={'line':line,'type':objectType}))
                raise InvalidCSVError()
            line += 1
        return line

    def _makeCompetencyGroup(self, container, data, line):
        if not data:
            self.errors.generic.append(_("Not enough data on line ${line}",
                                 mapping={'line':line}))
            raise InvalidCSVError()
        obj = Competencies(unicode(data[0]))
        chooser = INameChooser(container)
        name = chooser.chooseName('',obj)
        container[name] = obj
        return obj

    def _makeCompetency(self, container, data, line):
        if len(data) < 4:
            self.errors.generic.append(_("Not enough data on line ${line}",
                                 mapping={'line':line}))
            raise InvalidCSVError()

        compId, optional, globalId, title = [
            unicode(cell) for cell in data[:4]]

        if globalId not in self.global_comps:
            self.global_comps[globalId] = Competency(title)

        obj = CompetencyReference(
            compId, globalId, optionalFlag=optional)
        chooser = INameChooser(container)
        name = chooser.chooseName('',obj)
        container[name] = obj
        return obj

    def importFromCSV(self, csvdata):
        """Invoke importCompetencies while playing with BaseCSVImportView
        nicely.

        Currently sb.BaseCSVImportView expects ImporterClass.importFromCSV
        return True on success, False on error.  It would be nicer if it
        caught InvalidCSVErrors instead.  When this refactoring is performed,
        this method may be removed and importComeptencies can be renamed to
        importFromCSV.
        """
        try:
            if '\n' not in csvdata:
                self.errors.generic.append(_("No data provided"))
                raise InvalidCSVError()

            rows = self.parseCSVRows(csvdata.splitlines())
            self.importCompetencies(rows)
        except InvalidCSVError:
            return False
        else:
            return True


class CompetencyCSVImportView(BaseCSVImportView):
    """View for Competency CSV importer."""

    importer_class = CompetencyCSVImporter


class StudentCSVImporter(object):
    """A Student CSV importer, eSchoolPlus format."""

    _gender = {
        'm': _('male'),
        'f': _('female'),
        }

    def __init__(self, container, charset=None):
        self.container = container
        self.errors = ImportErrorCollection()
        if charset:
            self.charset = charset
        else:
            self.charset = "utf-8"
        try:
            term = getUtility(IDateManager).current_term
            year = ISchoolYear(term)
            groups = IGroupContainer(year)
        except:
            self.student_group = None
        else:
            self.student_group = groups.get('students')
        app = ISchoolToolApplication(None)
        self.demographics_container = app['schooltool.basicperson.demographics_data']

    def importFromCSV(self, csvdata):
        reader = csv.reader(StringIO(csvdata.strip()))
        lines_with_errors = {}
        for line_number, line in enumerate(reader):
            data = {}
            errors = self.getData(line, data)
            if errors:
                lines_with_errors[line_number+1] = errors
            else:
                self.process(data)
        if lines_with_errors:
            self.setErrors(lines_with_errors)
            return False
        return True

    def setErrors(self, lines_with_errors):
        for line_number, errors in sorted(lines_with_errors.items()):
            self.errors.fields.append(u". ".join([translate(e)
                                                  for e in errors]))
    
    def process(self, data):
        username = data['username']
        first_name = data['first_name']
        last_name = data['last_name']
        birth_date = data['birth_date']
        gender = data['gender']
        if username in self.container:
            new_student = False
            student = removeSecurityProxy(self.container[username])
            student.first_name = first_name
            student.last_name = last_name
        else:
            new_student = True
            student = BasicPerson(username, first_name, last_name)
            password = u'%02d%02d%02d' % (birth_date.month, birth_date.day,
                                          birth_date.year % 100)
            student.setPassword(password)
            self.container[username] = student
        student.birth_date = birth_date
        student.gender = gender
        if not new_student:
            notify(ObjectModifiedEvent(student))         
        if self.student_group is not None and \
           student not in self.student_group.members:
            self.student_group.members.add(student)

        demographics = self.demographics_container.get(student.username, None)
        if demographics is None:
            self.demographics_container[student.username] = demographics = PersonDemographicsData()

        try:
            demographics['ethnicity'] = data['ethnicity']
        except (KeyError,):
            pass
        try:
            demographics['grade_class'] = data['grade']
        except (KeyError,):
            pass
        try:
            demographics['state_report_id'] = data['state_report_id']
        except (KeyError,):
            pass

    def getData(self, line, data):
        errors = []
        columns = [column.decode(self.charset).strip() for column in line]
        if len(columns) < 4:
            errors.append(_("""Insufficient data provided."""))
            return errors           

        username = columns[0]
        first_name = columns[1]
        last_name = columns[2]
        birthdate = columns[3]

        if not username:
            errors.append(_('Student logins (student ids) may not be empty'))
            return errors

        if not first_name:
            errors.append(_('First names may not be empty'))
            return errors

        if not last_name:
            errors.append(_('Last names may not be empty'))
            return errors

        if not birthdate:
            errors.append(_('Birth dates may not be empty'))
            return errors

        try:
            # TODO: replace this minimalistic date parsing           
            parsed_birth_date = parse_date(birthdate)
        except (ValueError,):
            errors.append(
                _("Invalid date (${bdate}), expected yyyy-mm-dd",
                  mapping={'bdate': birthdate}))
            return errors

        try:
            gender_code = columns[4]
            if gender_code and gender_code.lower() not in self._gender:
                errors.append(
                    _("Invalid gender code (${code}), expected F or M",
                      mapping={'code': gender_code}))
                return errors
        except (IndexError,):
            gender_code = ''

        gender = self._gender.get(gender_code.lower())

        try:
            grade = columns[6]
            grade = int(grade)
        except (IndexError,):
            grade = None
        except (ValueError,):
            if grade and grade:
                errors.append(
                    _("Invalid grade (${grade}), expected an integer number",
                      mapping={'grade': grade}))
                return errors
            else:
                grade = None

        if username[:1] in '+@' or '/' in username:
            errors.append(
                _("Student logins cannot begin with '+' or '@' or contain '/'"))
            return errors

        try:
            code = columns[5].lower()
            ethnicity = _ethnicity_codes.get(code)
        except (IndexError,):
            ethnicity = None

        try:
            state_report_id = columns[7]
        except (IndexError,):
            state_report_id = None

        data['username'] = username
        data['first_name'] = first_name
        data['last_name'] = last_name
        data['birth_date'] = parsed_birth_date
        data['gender'] = gender
        data['ethnicity'] = ethnicity
        data['grade'] = grade
        data['state_report_id'] = state_report_id

        return errors


class StudentCSVImportView(BaseCSVImportView):
    """View for Student CSV importer."""

    importer_class = StudentCSVImporter


class SectionEnrollmentCSVImporter(object):
    """A Sections and Enrollment CSV importer."""

    def __init__(self, container, charset=None):
        self.container = container
        self.errors = ImportErrorCollection()
        if charset:
            self.charset = charset
        else:
            self.charset = "utf-8"
        app = ISchoolToolApplication(None)
        self.demographics_container = app['schooltool.basicperson.demographics_data']
        self.persons_container = app['persons']
        course_container = ICourseContainer(self.container)
        sy  = ISchoolYear(self.container)
        group_container = IGroupContainer(sy)
        self.courses = {}
        for course in course_container.values():
            course_id = course.course_id
            if course_id is not None:
                course_ids = [ID.strip() for ID in course_id.split(",")]
                for ID in course_ids:
                    self.courses[ID] = course
        self.teacher_group = None
        if 'teachers' in group_container:
            self.teacher_group = group_container['teachers']
        if 'students' in group_container:
            self.student_group = group_container['students']

    def setErrors(self, lines_with_errors):
        for line_number, errors in sorted(lines_with_errors.items()):
            self.errors.fields.append(u". ".join([translate(e)
                                                  for e in errors]))

    def importFromCSV(self, csvdata):
        reader = csv.reader(StringIO(csvdata.strip()))
        lines_with_errors = {}
        for line_number, line in enumerate(reader):
            data = {}
            errors = self.getData(line, data)
            if errors:
                lines_with_errors[line_number+1] = errors
            else:
                self.process(data)
        if lines_with_errors:
            self.setErrors(lines_with_errors)
            return False
        return True

    def process(self, data):
        course = data['course']
        course_id = data['course_id']
        description = data['description']
        building = data['building']
        marking_period = data['marking_period']
        start_period = data['start_period']
        section = data['section']
        student = data['student']
        teacher_first_name = data['teacher_first_name']
        teacher_last_name = data['teacher_last_name']
        teacher = data['teacher']

        if teacher is None:
            username = self.getTeacherUsername(teacher_first_name,
                                               teacher_last_name)
            last = teacher_last_name.title()
            first = teacher_first_name.title()
            teacher = BasicPerson(username, first, last)
            teacher.setPassword("teacher")
            self.persons_container[username] = teacher

        if self.teacher_group is not None and \
           teacher not in self.teacher_group.members:
            self.teacher_group.members.add(teacher)

        if self.student_group is not None and \
           student not in self.student_group.members:
            self.student_group.members.add(student)

        demographics = self.demographics_container.get(teacher.username, None)
        if demographics is None:
            self.demographics_container[teacher.username] = demographics = PersonDemographicsData()

        try:
            demographics['teacher_id'] = data['teacher_id']
        except (KeyError,):
            pass
        
        section = self.getSection(course, course_id, description, building,
                                  marking_period, start_period, section)

        if teacher not in section.instructors:
            section.instructors.add(teacher)

        if student not in section.members:
            section.members.add(student)

    def getData(self, line, data):
        errors = []
        columns = [column.decode(self.charset).strip() for column in line]

        if len(columns) < 10:
            errors.append(_("""Insufficient data provided."""))
            return errors

        course_id = columns[0]
        description = columns[1]
        building = columns[2]
        marking_period = columns[3]
        start_period = columns[4]
        section = columns[5]
        teacher_id = columns[6]
        teacher_first_name = columns[7]
        teacher_last_name = columns[8]
        student_id = columns[9]

        if not course_id:
            errors.append(_('Course IDs may not be empty'))
            return errors

        if not description:
            errors.append(_('Section descriptions may not be empty'))
            return errors

        if not building:
            errors.append(_('Building names may not be empty'))
            return errors

        if not marking_period:
            errors.append(_('Marking periods  may not be empty'))
            return errors

        if not start_period:
            errors.append(_('Start periods may not be empty'))
            return errors

        if not section:
            errors.append(_('Sections may not be empty'))
            return errors

        if not teacher_id:
            errors.append(_('Teacher IDs may not be empty'))
            return errors       

        if not teacher_first_name:
            errors.append(_('Teacher first name may not be empty'))
            return errors

        if not teacher_last_name:
            errors.append(_('Teacher last name may not be empty'))
            return errors

        if not student_id:
            errors.append(_('Student IDs may not be empty'))
            return errors

        course = self.getCourse(course_id)
        if course is None:
            errors.append(_('Provided Course ID was not found: '
                            '${ID}', mapping={'ID': course_id}))
            return errors

        try:
            teacher = self.getTeacher(teacher_id, teacher_first_name,
                                      teacher_last_name)
        except (Exception,), e:
            errors.append(e.message)
            return errors

        student = self.getStudent(student_id)
        if student is None:
            errors.append(_('Provided Student ID was not found: '
                            '${ID}', mapping={'ID': student_id}))
            return errors

        data['course_id'] = course_id
        data['description'] = description
        data['building'] = building
        data['marking_period'] = marking_period
        data['start_period'] = start_period
        data['section'] = section
        data['teacher_id'] = teacher_id
        data['teacher_first_name'] = teacher_first_name
        data['teacher_last_name'] = teacher_last_name
        data['student_id'] = student_id
        data['course'] = course
        data['teacher'] = teacher
        data['student'] = student

        return errors

    def getSection(self, course, course_id, description, building,
                   marking_period, start_period, section):
        title = self.buildSectionTitle(building, marking_period,
                                       start_period, section)
        description = self.buildSectionDescription(description, course_id)
        for section in course.sections:
            if title == section.title and description == section.description:
                return section
        section = Section(title=title, description=description)
        name = INameChooser(self.container).chooseName('', section)
        self.container[name] = section
        section.courses.add(removeSecurityProxy(course))
        return section

    def buildSectionTitle(self, building, marking_period, start_period,
                          section):
        return "%s, %s, %s, %s" % (building, marking_period, start_period,
                                   section)

    def buildSectionDescription(self, description, course_id):
        return "%s, %s" % (description, course_id)

    def getTeacherUsername(self, teacher_first_name, teacher_last_name):
        username = u"%s_%s" % (teacher_first_name.lower(),
                               teacher_last_name.lower())
        return username

    def getTeacher(self, teacher_id, teacher_first_name, teacher_last_name):
        username = self.getTeacherUsername(teacher_first_name,
                                           teacher_last_name)
        teacher = self.persons_container.get(username)
        if teacher is not None:
            demographics = self.demographics_container.get(teacher.username)
            if demographics is None:
                return teacher
            else:
                current_teacher_id = demographics.get('teacher_id')
                if current_teacher_id is None:
                    return teacher
                else:
                    if current_teacher_id != teacher_id:
                        raise Exception(u"There is another teacher with the "
                                        u"same username but with a different "
                                        u"Teacher ID: (%s, %s)" % (username,
                                                                   teacher_id))
                    else:
                        return teacher

    def getStudent(self, student_id):
        return self.persons_container.get(student_id)

    def getCourse(self, course_id):
        return self.courses.get(course_id)


class SectionEnrollmentCSVImportView(BaseCSVImportView):
    """View for Section and Enrollment CSV importer."""

    importer_class = SectionEnrollmentCSVImporter
