#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Virginia Competency views.

$Id:$
"""

import copy

from zope.container.interfaces import INameChooser
from zope.app.generations.utility import findObjectsProviding
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import adapts
from zope.formlib import form
from zope.interface import Interface, implements
from zope.publisher.browser import BrowserView
from zope.schema import TextLine, ValidationError
from zope.security.proxy import removeSecurityProxy
from zope.session.interfaces import ISession
from zope.viewlet.interfaces import IViewletManager
from zope.traversing.browser.absoluteurl import absoluteURL
import z3c.form
import z3c.form.validator

import schooltool.relationship
from schooltool.person.interfaces import IPerson
from schooltool.app.browser.app import BaseEditView
from schooltool.skin.containers import ContainerView
from schooltool.requirement.requirement import getRequirementKey
from schooltool.requirement.interfaces import IEvaluation
from schooltool.app.app import getSchoolToolApplication
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.course.interfaces import ICourse, ISection
from schooltool.securitypolicy.crowds import Crowd

from cando.virginia import interfaces
from cando.virginia.competency import buildCompetencyDict
from cando.virginia.competency import buildCompetencyDictByCourse
from cando.virginia.competency import Competencies
from cando.virginia.competency import CompetencyReference
from cando.virginia.competency import assignCoreCompetencies
from cando.virginia.competency import assignAdditionalCompetencies
from cando.virginia.competency import unassignAllCoreCompetencies
from cando.virginia.competency import unassignCompetencies
from cando.virginia.competency import getFlatCompetencies
from cando.virginia.scoresystem import UniversalScore
from cando.virginia.competency import getAdditionalCompetencies

from zope.i18nmessageid import MessageFactory

_ = MessageFactory('cando')


def dummy_getCompetencyFromGlobalId(globalId):
    pass


class ICompetencyViewletManager(IViewletManager):
    """Provides a viewlet hook for the virginia competency page."""


class CompetenciesEditView(BaseEditView):
    """View for editing Competencies."""

    __used_for__ = interfaces.ICompetencies


class VirginiaCompetencyEditView(BaseEditView):
    """View for editing Competencies."""

    __used_for__ = interfaces.ICompetencyReference


class VirginiaCompetencyView(BrowserView):

    def __init__(self, *args, **kw):
        super(VirginiaCompetencyView, self).__init__(*args, **kw)

        self.SOLCorrelations = []
        sols = removeSecurityProxy(self.context).SOLCorrelations
        for category, ids in sorted(sols.items()):
            self.SOLCorrelations.append({
                'category': category,
                'correlations': u', '.join(ids),
                })
        self.modified = []
        modified = removeSecurityProxy(self.context).modified
        timestamp_format = "%Y-%m-%d %H:%M:%S"
        if modified is not None:
            for timestamp, old_title, new_title in sorted(modified,
                                                          key=lambda x:x[0]):
                self.modified.append({
                    'timestamp': timestamp.strftime(timestamp_format),
                    'old_title': old_title,
                    'new_title': new_title
                    })


class ImageAction(form.Action):

    image = "/@@/competencyedit.png"
    title = ""
    def render(self):
        if not self.available():
            return ''
        return ('<input type="submit" id="%s" name="%s" value="  "'
                ' class="button" '
                ' style="background: url(\'%s\');'
                       'border: 0px; height: 25px; width: 25px;"'
                ' title="%s"/>' %
                   (self.__name__, self.__name__, self.image, self.title)
                    )
class EditAction(ImageAction):
    image = "/@@/competencyedit.png"
    title = "Edit"
class DeleteAction(ImageAction):
    image = "/@@/competencydelete.png"
    title = "Delete"
class OverrideAction(ImageAction):
    image = "/@@/override.png"
    title="Override"
class SelectAction(form.Action):
    def render(self):
        if not self.available():
            return ''
        return ('<input type="submit" id="%s" name="%s" value="%s"'
                ' class="button" title="%s"/>' %
                (self.__name__, self.__name__, self.label, self.label))


class SelectedCompetenciesFormView(form.EditForm):
    """Form for managing selected competencies."""

    form_fields = form.Fields()
    prefix = "selected_comps"
    template = ViewPageTemplateFile("templates/selectedcompetencies.pt")

    actions = form.Actions(
        form.Action("Clear", success="handle_clear_action"),
        form.Action("Link Here", success="handle_link_action"))

    def handle_clear_action(self, action, data):
        sessionData = ISession(self.request)['cando.virginia.competency']
        sessionData['selected'] = []

    def handle_link_action(self, action, data):
        sessionData = ISession(self.request)['cando.virginia.competency']
        selected = sessionData.get('selected',[])
        for comp in selected:
            chooser = INameChooser(self.context)
            name = chooser.chooseName(comp.__name__, comp)
            self.context[name] = comp
        self.status = "Selected Competencies Linked"

    def setUpData(self):
        sessionData = ISession(self.request)['cando.virginia.competency']
        self.selectedCompetencies = sessionData.get('selected',[])

    def __call__(self):
        self.update()
        self.setUpData()
        return self.render()


class BaseCompetencyTreeNodeView(form.EditFormBase):
    """Base view for competencies."""
    form_fields = form.Fields(interfaces.ICompetencies,
                              for_display=True).select('title')
    subForm = None
    subFormActions = form.Actions(
        EditAction("Edit", success="handle_subform_action",
                            data="EditForm"))
    perActions = form.Actions()
    regActions = form.Actions(
        DeleteAction("Delete", success="handle_delete_action"),
        SelectAction("Select", success="handle_select_action",
                               condition="condition_select_action"),
        SelectAction("Unselect", success="handle_unselect_action",
                               condition="condition_unselect_action"))

    actions = property(lambda self: self.subFormActions + self.perActions)

    def handle_subform_action(self, action, data):
        self.subForm = self.subForms.get(action.data)

    def handle_delete_action(self, action, data):
        try:
            if self.context.__parent__:
                del self.context.__parent__[self.context.__name__]
        except KeyError:
            pass

    def handle_select_action(self, action, data):
        session = ISession(self.request)
        sessionData = session['cando.virginia.competency']
        if 'selected' not in sessionData.keys():
            sessionData['selected'] = []
        sessionData['selected'].append(removeSecurityProxy(self.context))

    def handle_unselect_action(self, action, data):
        selected = ISession(self.request)['cando.virginia.competency'].get('selected', [])
        if self.context in selected:
            selected.remove(self.context)

    def condition_select_action(self, action):
        self.setUpSessionVariables()
        if self.sessionSelect:
                return False
        return True
    def condition_unselect_action(self, action):
        self.setUpSessionVariables()
        if self.sessionSelect:
                return True
        return False

    def handle_persistent_action(self, action, data):
        form = self.persistentForms.get(action.data)
        if form is not None:
            form.show = True

    def setUpPersistentForms(self):
        self.persistentForms = {}

    def setUpSubForms(self):
        self.subForms = {}
        self.subForms['EditForm'] = BaseTreeNodeEditView(self.context,
                                                         self.request,
                                                         self.form_fields)
    def setUpSessionVariables(self):
        session = ISession(self.request)
        sessionData = session['cando.virginia.competency']
        if 'selected' not in sessionData.keys():
            sessionData['selected'] = []
        if self.context in sessionData['selected']:
            self.sessionSelect = True
        else:
            self.sessionSelect = False

    def render(self):
        self.setPrefix('comps'+str(hash(getRequirementKey(self.context))))
        self.setUpSubForms()
        self.setUpPersistentForms()
        super(BaseCompetencyTreeNodeView, self).update()
        return super(BaseCompetencyTreeNodeView, self).render()

    def update(self, subForms=True):
        self.setPrefix('comps'+str(hash(getRequirementKey(self.context))))
        self.setUpSubForms()
        if subForms:
            for subForm in self.subForms.values():
                subForm.update()
        self.setUpPersistentForms()
        for form in self.persistentForms.values():
            form.update()
        super(BaseCompetencyTreeNodeView, self).update()

    def __call__(self):
        self.update(subForms=False)
        return self.render()


class CompetenciesTreeNodeView(BaseCompetencyTreeNodeView):
    """View for competencies tree nodes."""

    form_fields = form.Fields(interfaces.ICompetencies,
                              for_display=True).select('title')
    template = ViewPageTemplateFile("templates/competenciestreenode.pt")

    def setUpSubForms(self):
        super(CompetenciesTreeNodeView, self).setUpSubForms()
        self.subForms['AddGroupForm'] = CompetenciesTreeNodeAddView(
                                                               self.context,
                                                               self.request)
        self.subForms['AddCompForm'] = VirginiaCompetencyTreeNodeAddView(
                                                               self.context,
                                                               self.request)

    def setUpPersistentForms(self):
        super(CompetenciesTreeNodeView, self).setUpPersistentForms()


class BaseTreeNodeEditView(form.EditFormBase):
    """View editing nodes in the tree."""
    template = ViewPageTemplateFile("templates/editformlib.pt")
    def __init__(self, context, request, form_fields=()):
        super(BaseTreeNodeEditView, self).__init__(context, request)
        self.form_fields = [copy.copy(field) for field in form_fields]
        for field in self.form_fields:
            field.for_display = False
    def update(self):
        self.setPrefix('comps_edit'+str(hash(getRequirementKey(self.context))))
        super(BaseTreeNodeEditView, self).update()


class CompetenciesTreeNodeAddView(form.AddFormBase):
    """View for adding competency groups."""
    form_fields = form.Fields(interfaces.ICompetencies).select('title')
    template = ViewPageTemplateFile("templates/editformlib.pt")
    def create(self, data):
        return Competencies(data.get('title',''))
    def add(self, obj):
        name = INameChooser(self.context).chooseName('',obj)
        self.context[name] = obj
        self._finished_add = True
        return obj
    def update(self):
        self.setPrefix('comps_add'+str(hash(getRequirementKey(self.context))))
        super(CompetenciesTreeNodeAddView, self).update()


class VirginiaCompetencyTreeNodeAddView(form.AddFormBase):
    """View for adding competency groups."""
    form_fields = form.Fields(interfaces.ICompetencyReference).select(
                                          'id','optionalFlag','title')
    template = ViewPageTemplateFile("templates/editformlib.pt")
    def create(self, data):
        return CompetencyReference(data.get('id',''),
                                   data.get('title',''),
                                   data.get('optionalFlag'))
    def add(self, obj):
        name = INameChooser(self.context).chooseName('',obj)
        self.context[name] = obj
        self._finished_add = True
        return obj
    def update(self):
        self.setPrefix('vacomp_add'+str(hash(getRequirementKey(self.context))))
        super(VirginiaCompetencyTreeNodeAddView, self).update()


class VirginiaCompetencyTreeNodeView(BaseCompetencyTreeNodeView):
    """View for virginia competency tree nodes."""

    form_fields = form.Fields(interfaces.ICompetencyReference,
                              for_display=True).select(
                                                   "id", "optionalFlag", "title", "globalId")
    prefix = "competency"
    template = ViewPageTemplateFile("templates/virginiacompetencytreenode.pt")


class VirginiaCompetencyTreeNodeEditView(form.SubPageEditForm):
    """View for editing virginia competency tree nodes."""

    template = ViewPageTemplateFile("templates/virginiacompetencytreenode.pt")


class CompetenciesView(ContainerView):
    """A Requirement view."""

    __used_for__ = interfaces.ICompetencies

    index_title = _("Competencies index")
    add_title = _("Add a new Competency")
    add_url = "+/addCompetency.html"

    def __init__(self, context, request, depth=None):
        super(CompetenciesView, self).__init__(context, request)
        self.depth = depth
        if self.depth is None:
            self.depth = int(request.get('DEPTH', 3))

        sessionData = ISession(self.request)['cando.virginia.competency']
        self.selectedCompetencies = sessionData.get('selected',[])

    def _search(self, searchstr, context):
        results = []
        for item in context.values():
            if searchstr.lower() in item.title.lower():
                results.append(item)
            results += self._search(searchstr, item)
        return results

    def listContentInfo(self):
        children = []
        if self.depth < 1:
            return []
        for child in self.batch:
            if schooltool.requirement.interfaces.IRequirement.providedBy(child):
                info = {}
                info['child'] = child
                thread = CompetenciesView(child, self.request, self.depth-1)
                info['thread'] = thread.subthread()
                children.append(info)
        return children

    subthread = ViewPageTemplateFile('templates/subthread.pt')


class CompetenciesClassicView(BrowserView):
    """A view for the classic version of the competencies
    editor view.
    """

    def update(self):
        pass

    def _getLevel (self, competency):
        i = 0
        while not ISchoolToolApplication.providedBy(competency):
            competency = competency.__parent__
            i += 1
        return i - 1

    def _getData(self, competency, sub=False):
        isGroup = interfaces.ICompetencies.providedBy(competency)
        data = {'isHeader':isGroup and not sub,
                'subGroup':isGroup and sub,
                'objectURL':absoluteURL(competency,self.request),
                'id':'',
                'optionalFlag':'',
                'name':competency.__name__,
                'level':self._getLevel(competency),
                'object':competency}
        if interfaces.ICompetencyReference.providedBy(competency):
            data['title'] = interfaces.ICompetency(competency).title
            data['id'] = competency.id
            data['globalId'] = competency.globalId
            data['optionalFlag'] = competency.optionalFlag
        else:
            data['title'] = competency.title
        return data

    def isCourseOrGroup(self, competency):
        return self._getLevel(competency) > 1

    def sortKey(self, competency):
        comprefs = findObjectsProviding(competency,
                                        interfaces.ICompetencyReference)
        ids = sorted([compref.id for compref in comprefs])
        if ids:
            return ids[0]
        else:
            return ''

    def getRows(self):
        rows = []
        if self.isCourseOrGroup(self.context):
            values = sorted(self.context.values(), key=self.sortKey)
        else:
            values = sorted(self.context.values(), key=lambda x:x.title)
        for competency in values:
            rows.append(self._getData(competency))
            if interfaces.ICompetencies.providedBy(competency):
                if self.isCourseOrGroup(competency):
                    rows += [self._getData(comp, sub=True)
                             for comp in sorted(competency.values(),
                                                key=self.sortKey)]
                else:
                    rows += [self._getData(comp, sub=True)
                             for comp in sorted(competency.values(),
                                                key=lambda x:x.title)]
        return rows


class ViewCompetencies(BrowserView):

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        self.comps = buildCompetencyDict(removeSecurityProxy(self.context))
        self.additionalComps = {}
        self.getSectionAdditionalComps()
        self.getCoursesAdditionalComps()

    def groups(self):
        core_groups = []
        additional_groups = []
        for group_title in self.comps:
            if self.comps[group_title]:
                firstComp = self.comps[group_title][0]
                if self.isAdditionalCompetency(firstComp):
                    additional_groups.append(group_title)
                else:
                    core_groups.append(group_title)
        return sorted(core_groups, key=self.sortKey) + \
               sorted(additional_groups, key=self.sortKey)

    def sortKey(self, group_title):
        ids = sorted([compref.id for compref in self.comps[group_title]])
        if ids:
            return ids[0]
        else:
            return ''

    def getSectionAdditionalComps(self):
        section = self.getSection()
        for comp in getAdditionalCompetencies(section):
            self.additionalComps[comp] = None

    def getCoursesAdditionalComps(self):
        courses = self.getCourses()
        for course in courses:
            for comp in getAdditionalCompetencies(course):
                self.additionalComps[comp] = None

    def getSection(self):
        return removeSecurityProxy(self.context)

    def getCourses(self):
        if ICourse.providedBy(self.context):
            return [removeSecurityProxy(self.context),]
        return list(self.getSection().courses)
        
    def isAdditionalCompetency(self, comp):
        return comp in self.additionalComps

    def getCompIdClass(self, comp):
        if self.isAdditionalCompetency(comp):
            return "additional-comp"
        return "%s-comp" % comp.optionalFlag


class RemoveCompetencies(BrowserView):

    context_type_name = 'course'

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        self.compDict = buildCompetencyDictByCourse(
            removeSecurityProxy(self.context))
        self.compDictForDisplay = sorted(self.compDict)
        self.getFields()
        self.removed = []
        self.skipped = []
        self.update()

    def getFields(self):
        self.course = self.request.get('course')
        self.cgroup = self.request.get('cgroup')
        self.comps = self.request.get('competency')

    def update(self):
        if 'Done' in self.request:
            self.request.response.redirect('index.html')
        if 'remove_competencies' in self.request:
            comps = self.request.get('competency')
            if type(comps) == list:
                self.removeCompetencies(comps)
            if type(comps) == str or type(comps) == unicode:
                self.removeCompetencies([comps])
        if 'remove_cgroup' in self.request:
            self.removeCGroup()
        if 'remove_course' in self.request:
            self.removeCourse()
        #updating competency dict to reflect changes
        self.compDict = buildCompetencyDictByCourse(
            removeSecurityProxy(self.context))

    def rememberRemovedComps(self, compref_tree_node):
        comps_to_remove = getFlatCompetencies(compref_tree_node)
        assigned_comps = self.compDict.values()
        self.removed.extend([comp for comp in comps_to_remove
                             if comp not in assigned_comps])
        self.skipped.extend([comp for comp in comps_to_remove
                             if comp in assigned_comps])

    def removeCompetencies(self, selected):
        context = removeSecurityProxy(self.context)
        if self.compDict[self.course].has_key(self.cgroup):
            comps_by_globalId = {}
            for comp in self.compDict[self.course][self.cgroup]:
                comps_by_globalId[comp.globalId] = comp
            comps_to_remove = [comps_by_globalId[globalId]
                               for globalId in selected
                               if globalId in comps_by_globalId]
            for comp in comps_to_remove:
                self.rememberRemovedComps(comp)
                unassignCompetencies(comp, context)

    def removeCGroup(self):
        context = removeSecurityProxy(self.context)
        if self.compDict[self.course].has_key(self.cgroup):
            for comp in self.compDict[self.course][self.cgroup]:
                self.rememberRemovedComps(comp)
                unassignCompetencies(comp, context)

    def removeCourse(self):
        context = removeSecurityProxy(self.context)
        if self.compDict.has_key(self.course):
            for cgroup in self.compDict[self.course].values():
                for competency in cgroup:
                    self.rememberRemovedComps(competency)
                    unassignCompetencies(competency, context)


class RemoveSectionCompetencies(RemoveCompetencies):

    context_type_name = 'section'


class RemoveSkilldriverCompetencies(RemoveCompetencies):

    context_type_name = 'assignment'

    def update(self):
        if 'Done' in self.request:
            self.request.response.redirect('index.html')
        else:
            super(RemoveSkilldriverCompetencies, self).update()


class DeleteCompetenciesView(BrowserView):

    redirected = False

    def __init__(self, context, request):
       	BrowserView.__init__(self, context, request)
        self.complist = getFlatCompetencies(
            removeSecurityProxy(self.context))

    def update(self):
        if 'Cancel' in self.request:
            self.redirectToNodeIndex(self.context)
        if 'Delete' in self.request:
            for comp in self.complist:
                schooltool.relationship.unrelateAll(comp)
            self.deleteAndRedirect(removeSecurityProxy(self.context))

    def redirectToNodeIndex(self, node):
        self.request.response.redirect(
            absoluteURL(node, self.request))
        self.redirected = True

    def deleteAndRedirect(self, node):
        node_parent = node.__parent__
        if interfaces.ICompetencies.providedBy(node_parent):
            del node_parent[node.__name__]
            if not len(node_parent):
                self.deleteAndRedirect(node_parent)
            else:
                self.redirectToNodeIndex(node_parent)
        else:
            for key in list(node.keys()):
                del node[key]
            self.redirectToNodeIndex(node)

    def __call__(self):
        self.update()
        if self.redirected:
            return ''
        else:
            return self.index()


class SuffixValidationError(ValidationError):

    def doc(self):
        return str(self)


class IDeprecateForm(Interface):

    suffix = TextLine(
        title=_(u"Suffix"),
        description=_(u"2-digit suffx for competency ids"),
        required=True)


class SuffixValidator(z3c.form.validator.SimpleFieldValidator):

    def validate(self, suffix):
        try:
            if suffix is None or not suffix.strip() or len(suffix) != 2:
                raise ValueError
            suffix = int(suffix)
            if suffix < 0:
                raise ValueError
        except (ValueError,):
            raise SuffixValidationError(_("Please enter a 2-digit suffix "
                                          "for last year's competencies"))


z3c.form.validator.WidgetValidatorDiscriminators(SuffixValidator,
                                                 field=IDeprecateForm['suffix'])


class DeprecateFormAdapter(object):

    implements(IDeprecateForm)
    adapts(interfaces.ICompetencies)

    def __init__(self, context):
        self.context = context


class DeprecateCompetenciesView(z3c.form.form.Form):

    template = ViewPageTemplateFile('templates/deprecate_competencies.pt')
    fields = z3c.form.field.Fields(IDeprecateForm)
    label = _('Deprecate Selected Competencies')

    @z3c.form.button.buttonAndHandler(_('Deprecate'))
    def handle_ok_action(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = _('There were errors')
            return
        competencies = findObjectsProviding(self.context,
                                            interfaces.ICompetencyReference)
        suffix = int(data['suffix'])
        for competency in competencies:
            if not competency.deprecated:
                competency.id = "%s-%02d" % (competency.id, suffix,)
                competency.deprecated = True
                competency.optionalFlag = 'optional'
        url = absoluteURL(self.context ,self.request)
        self.request.response.redirect(url)

    @z3c.form.button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context ,self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(DeprecateCompetenciesView, self).updateActions()
        self.actions['deprecate'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class AssignCompetenciesView(BrowserView):

    context_type_name = 'course'

    def __init__(self, context, request):
       	BrowserView.__init__(self, context, request)
        self.collectCompetencies()
        self.getFields()
        self.added = []
        self.skipped = []
        self.update()

    def collectCompetencies(self):
        self.globalComps = removeSecurityProxy(interfaces.ICompetencies(
            getSchoolToolApplication()))

    def getFields(self):
        self.area = self.request.get('area')
        self.course = self.request.get('course')
        self.group = self.request.get('group')

    def redirectToIndex(self):
        self.request.response.redirect('view_competencies.html')

    def update(self):
        if 'Done' in self.request:
            self.redirectToIndex()
        if 'add_competencies' in self.request:
            comps = self.request.get('competency')
            if type(comps) == list:
                self.addCompetencies(comps)
            if type(comps) == str or type(comps) == unicode:
                self.addCompetencies([comps])
        if 'add_group' in self.request:
            self.addGroup(self.group)
        if 'add_course' in self.request:
            self.addCourse()

    def rememberAddedComps(self, comp_ref_tree_node):
        context = removeSecurityProxy(self.context)
        comps_to_assign = getFlatCompetencies(
            removeSecurityProxy(comp_ref_tree_node))
        assigned_comps = interfaces.ICompetencyList(context)
        self.added.extend([comp for comp in comps_to_assign
                           if comp not in assigned_comps])
        self.skipped.extend([comp for comp in comps_to_assign
                             if comp in assigned_comps])

    def addCompetencies(self, selected):
        context = removeSecurityProxy(self.context)
        if self.getCompetencies():
            available = self.globalComps[self.area][self.course][self.group]
            for comp in selected:
                if comp in available:
                    self.rememberAddedComps(available[comp])
                    assignAdditionalCompetencies(available[comp], context)

    def addGroup(self, group):
        context = removeSecurityProxy(self.context)
        if self.getGroups():
            av_groups = self.globalComps[self.area][self.course]
            if group in av_groups:
                self.rememberAddedComps(av_groups[group])
                assignAdditionalCompetencies(av_groups[group], context)

    def addCourse(self):
        context = removeSecurityProxy(self.context)
        # only one set (full course) of core competencies is allowed
        unassignAllCoreCompetencies(context)
        if self.getCourses():
            av_courses = self.globalComps[self.area]
            if self.course in av_courses:
                # if core competencies were assigned as additional before,
                # remove them
                unassignCompetencies(av_courses[self.course], context)
                # assign new core competencies
                self.rememberAddedComps(av_courses[self.course])
                assignCoreCompetencies(av_courses[self.course], context)

    def getAreas(self):
        return sorted(self.globalComps.values(), key=lambda x:x.title)

    def getCourses(self):
        if self.area and self.globalComps.has_key(self.area):
            return sorted(self.globalComps[self.area].values(),
                          key=lambda x:x.title)

    def getGroups(self):
        if self.area and self.globalComps.has_key(self.area):
            areaObj = self.globalComps[self.area]
            if self.course and areaObj.has_key(self.course):
                return sorted(areaObj[self.course].values(),
                              key=lambda x:x.title)

    def getCompetencies(self):
        if self.area and self.globalComps.has_key(self.area):
            areaObj = self.globalComps[self.area]
            if self.course and areaObj.has_key(self.course):
                groupObj = areaObj[self.course]
                if self.group and groupObj.has_key(self.group):
                    return sorted(groupObj[self.group].values(),
                                  key=lambda x:x.competency.title)


class AssignSectionCompetenciesView(AssignCompetenciesView):

    context_type_name = 'section'

    def addCourse(self):
        context = removeSecurityProxy(self.context)
        if self.getCourses():
            av_courses = self.globalComps[self.area]
            if self.course in av_courses:
                self.rememberAddedComps(av_courses[self.course])
                assignAdditionalCompetencies(av_courses[self.course], context)


class AssignSkillDriverCompetenciesView(AssignSectionCompetenciesView):

    context_type_name = 'assignment'

    def collectCompetencies(self):
        self.compDict = buildCompetencyDict(removeSecurityProxy(
            self.context.__parent__.__parent__))

    def redirectToIndex(self):
        self.request.response.redirect('index.html')

    def add(self, comps):
        context = removeSecurityProxy(self.context)
        assigned_global_ids = [compref.globalId
                               for compref in context.getCompetencies()]
        for comp in comps:
            if comp.globalId not in assigned_global_ids:
                assignAdditionalCompetencies(comp, context)
                self.added.append(comp)
            else:
                self.skipped.append(comp)

    def addCompetencies(self, selected):
        to_add = []
        if self.getCompetencies():
            for comp in selected:
                for competency in self.compDict[self.group]:
                    if competency.__name__ == comp:
                        to_add.append(competency)
        self.add(to_add)

    def addGroup(self, group):
        if self.getGroups():
            av_groups = self.compDict
            if group in av_groups:
                self.add(av_groups[group])

    def getGroups(self):
        core_groups = []
        additional_groups = []
        self.additionalComps = self.getAdditionalComps()
        for group in self.compDict:
            if self.compDict[group]:
                firstComp = self.compDict[group][0]
                if self.isAdditionalCompetency(firstComp):
                    additional_groups.append(group)
                else:
                    core_groups.append(group)
        return sorted(core_groups) + sorted(additional_groups)

    def getAdditionalComps(self):
        result = {}
        if ISection.providedBy(self.context.__parent__.__parent__):
            section = removeSecurityProxy(self.context.__parent__.__parent__)
            for comp in getAdditionalCompetencies(section):
                result[comp] = None
            for course in section.courses:
                for comp in getAdditionalCompetencies(course):
                    result[comp] = None
        if ICourse.providedBy(self.context.__parent__.__parent__):
            course = removeSecurityProxy(self.context.__parent__.__parent__)
            for comp in getAdditionalCompetencies(course):
                result[comp] = None
        return result
        
    def isAdditionalCompetency(self, comp):
        return comp in self.additionalComps

    def getCompetencies(self):
        groupObj = self.compDict
        if self.group and groupObj.has_key(self.group):
            return sorted(groupObj[self.group],
                          key=lambda x:x.competency.title)


class ScoreSystemView(form.PageDisplayForm):

    form_fields = form.Fields(interfaces.IUniversalScoreSystem).select(
                    'title', 'description')
    template = ViewPageTemplateFile("templates/scoresystem.pt")


class ScoreSystemEditView(form.EditForm):
    form_fields = form.Fields(interfaces.IUniversalScoreSystem).select(
                     'title', 'description')
    prefix = "score_system"
    template = ViewPageTemplateFile("templates/editscoresystem.pt")

    actions = form.Actions(
         form.Action('Delete', success='handle_delete_action'),
         form.Action('Apply', success='handle_edit_action'))

    def score_id(self, title):
	return str(hash(title))

    def handle_edit_action(self, action, data):
        form.applyChanges(self.context, self.form_fields, data)
        minPassing = self.request.get('score.minPassing')
        if minPassing:
            self.context.minPassingScore = removeSecurityProxy(
                self.context.fromUnicode(minPassing))
        else:
            self.context.minPassingScore = None

    def handle_delete_action(self, action, data):
        for score in self.context.scores:
            if 'delete.'+ self.score_id(score.title) in self.request.keys():
                self.context.removeScore(score)

    def setUpWidgets(self, ignore_request=False):
        super(ScoreSystemEditView, self).setUpWidgets(
                           ignore_request=ignore_request)
        self.widgets['title'].displayWidth = 40
        self.widgets['description'].width = 40
        self.widgets['description'].height = 2

    def addForm(self):
        return UniversalScoreAddForm(self.context, self.request)


class UniversalScoreView(form.SubPageDisplayForm):
    form_fields = form.Fields(interfaces.IUniversalScore)


class UniversalScoreEditView(form.SubPageEditForm):
    form_fields = form.Fields(interfaces.IUniversalScore).select(
        'title','value','weight')
    prefix="score"
    actions = form.Actions(
         form.Action('Cancel'),
         form.Action('Apply', success='handle_edit_action'))

    def handle_edit_action(self, action, data):
        form.applyChanges(self.context, self.form_fields, data)

    def update(self):
        self.setPrefix('score.%s' % hash(self.context))
        super(UniversalScoreEditView, self).update()


class UniversalScoreAddForm(form.AddFormBase):
    form_fields = form.Fields(interfaces.IUniversalScore).select(
        'title','value','weight')
    prefix="addscore"

    template = ViewPageTemplateFile("templates/addscore.pt")
    def create(self, data):
        return UniversalScore(data.get('title',''),
                              int(data.get('value','')),
                              float(data.get('weight','')))
    def add(self, obj):
        self.context.addScore(obj)
        self._finished_add = True
        return obj

    def nextURL(self):
        return self.request.URL


class EvaluationEvaluatorView(object):

    def __init__(self, context, request):
        self.context = removeSecurityProxy(context)
        self.request = request

    def __call__(self):
        recievedScore = self.request.get('value',u'')
        if recievedScore == u'-':
            recievedScore = u''
        score = self.context.requirement.scoresystem.fromUnicode(recievedScore)
        if self.context.value != score:
            self.context.value = score
        return recievedScore


class EvaluationEvaluateeCrowd(Crowd):
    """Crowd that contains only the evaluatee of the evaluation."""

    # context assumed to be an evaluation
    def __init__(self, context):
        if not IEvaluation.providedBy(context):
            raise TypeError("Expected a provider of IEvaluation but got %s instead", context)
        super(EvaluationEvaluateeCrowd, self).__init__(context)

    def contains(self, principal):
        person = IPerson(principal, None)
        return person == self.context.evaluatee


class DemographicsView(BrowserView):
    """Demographics view for testing purposes."""

    @property
    def info(self):
        return interfaces.IStudentDemographics(self.context)

