#
# CanDo - Web based competency tracking system.
# Copyright (C) 2006 Arlington Public Schools, Arlington, VA and The
# Shuttleworth Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Virginia ScoreSystem Implementation

$Id$
"""


from persistent import Persistent
import persistent.list
from zope.interface import implements
import zope.schema
from zope.schema.fieldproperty import FieldProperty
from zope import annotation
from zope.container import contained
from schooltool.requirement import scoresystem
from schooltool.requirement.scoresystem import UNSCORED
from cando.virginia import interfaces
from zope.i18nmessageid import MessageFactory

_ = MessageFactory('cando')

class UniversalScoreSystem(scoresystem.DiscreteValuesScoreSystem,
                           Persistent,
                           contained.Contained):

    implements(interfaces.IUniversalScoreSystem,
               zope.container.interfaces.IContained,
               annotation.interfaces.IAttributeAnnotatable)

    # See interfaces.IDiscreteValuesScoreSystem
    scores = []
    minPassingScore = None
    bestScore = None
    title = u''
    description = u''

    def __init__(self, title=None, description=None,
                 scores=None, bestScore=None, minPassingScore=None):
        self.title = title
        self.description = description
        self.scores = scores or persistent.list.PersistentList()
        self.bestScore = bestScore
        self.minPassingScore = minPassingScore

    def addScore(self, score):
        self.scores.append(score)
        self.scores.sort(lambda a,b: cmp(a.value,b.value))
        self.bestScore = self.scores[-1]

    def removeScore(self, score):
        if score not in self.scores:
            return
        self.scores.remove(score)

    def isPassingScore(self, score):
        """See interfaces.IScoreSystem"""
        if score is UNSCORED:
            return None
        if self.minPassingScore is None:
            return None
        return score.value >= self.minPassingScore.value

    def isValidScore(self, score):
        """See interfaces.IScoreSystem"""
        return score in self.scores + [UNSCORED]

    def getBestScore(self):
        """See interfaces.IScoreSystem"""
        return self.bestScore

    def fromUnicode(self, rawScore):
        """See interfaces.IScoreSystem"""
        if rawScore in ['','-']:
            return UNSCORED
        numScore = None
        try: # we have an integer in string form
            numScore = int(rawScore)
        except ValueError:
            try: # we have a float in string form
                numScore = int(float(rawScore))
            except ValueError: # we have the title of a score
                pass
        for score in self.scores:
            if int(score.value) == numScore:
                return score
            elif score.title == rawScore:
                return score
        raise zope.schema.ValidationError(
            _("'%s' is not a valid score." % rawScore))

    def getNumericalValue(self, score):
        """See interfaces.IScoreSystem"""
        if score is UNSCORED:
            return None
        return score.value

    def getFractionalValue(self, score):
        """See interfaces.IScoreSystem"""
        # get maximum and minimum score to determine the range
        maximum = self.getBestScore().value
        minimum = sorted(self.scores, cmp=lambda a,b: cmp(a.value, b.value))[0].value
        # normalized numerical score
        value = self.getNumericalValue(score) - minimum
        return float(value) / (maximum - minimum)


class UniversalScore(Persistent):

    implements(interfaces.IUniversalScore)

    title = FieldProperty(interfaces.IUniversalScore['title'])
    value = FieldProperty(interfaces.IUniversalScore['value'])
    weight = FieldProperty(interfaces.IUniversalScore['weight'])

    def __init__(self, title, value, weight=1.0):
        self.title = title
        self.value = value
        self.weight = weight

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __repr__(self):
        return '<%s %r %r weight=%r>' % (self.__class__.__name__, self.title,
                                  self.value, self.weight)


def addUniversalScoreSystemToApplication(event):
    uni = UniversalScoreSystem(_('Universal Score System'),
                               _('The score system used by all competencies.'))
    event.object['scoresystem'] = uni
    site = event.object.getSiteManager()
    site.registerUtility(uni, interfaces.IUniversalScoreSystem)
